// Copyright 1998-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniSCSI/OSCSI2CDRomUnit.h>

#import <OmniSCSI/OSCSICDTableOfContents.h>
#import <OmniSCSI/OSCSIGeneric.h>
#import <OmniSCSI/OSCSIResponse.h>

#import <dev/scsireg.h>

@interface OSCSI2CDRomUnit (PrivateAPI)
- (void) _readSubcodeStatus;
- (void) _readControlPage;
@end

@implementation OSCSI2CDRomUnit

+ (NSString *) displayName;
{
    return @"SCSI-2 CDROM";
}

- (void) dealloc;
{
    [_tableOfContents release];
    [super dealloc];
}

- (id <OSCSICDRecorder>) recorderUnit;
{
    // By default, we don't support writing
    return nil;
}

- (void) mediaInserted;
{
    [super mediaInserted];
    [self tableOfContents];
}

- (void) mediaRemoved;
{
    [super mediaRemoved];

    [_tableOfContents release];
    _tableOfContents = nil;

    _currentLBA = 0;
    _currentStatus = OSCSICDAudioIdleStatus;

    _currentTrack = 0;
    _currentIndex = 0;

    [[NSNotificationCenter defaultCenter] postNotificationName: OSCSICDAudioUnitTableOfContentsChangedNotification
                                                        object: self];
}

- (void) updateStatus;
{
    if (![self isOpen])
        [self open];
    
    [self testUnitReady];
    [self _readSubcodeStatus];
    [self _readControlPage];
}

- (BOOL) mediaPresent;
{
    return _mediaPresent;
}

- (void) unloadMedia;
{
    [self stopPlaying];
    [self eject];
}

- (OSCSICDTableOfContents *) tableOfContents;
{
    OSCSICDTableOfContentsEntry  *descriptions     = NULL;
    unsigned int                  descriptionIndex, descriptionCount = 0;
    NSMutableData                *tocData = nil;
    OSCSIResponse                *response = nil;
    unsigned int                  length, retryCount = 3;
    unsigned char                *bytes;


    if (_tableOfContents) {
        return _tableOfContents;
    }

    if (![self mediaIsPresent]) {
        // Don't ask, since if we are using the raw device, it will ask us to insert media.
        return nil;
    }

    if (![self isOpen])
        [self open];
    
    // The header is 4 bytes (TOC data length MBS, LSB, first track number, last track number).
    // The last two bytes are included in the TOC data length.  Each track consumes 8 bytes.
    // With at least *some* ATAPI devices, if you ask for more data than the TOC actually takes,
    // no bytes will be returned (thanks, guys).  Other ATAPI devices, if you only ask for 4 bytes
    // will return four bytes and not fill in the full length in the header.  Idiots.  Under SCSI,
    // all the devices I've seen will simply ignore the extra space.
    // So, what we'll do is start out by reading the maximum amount of data.  If we get a full
    // TOC, so be it.  If we don't, then we'll start out at 4 bytes and try to read just that.
    // That should give us the actual length that we should use (after adding two since the
    // first two bytes aren't included in the returned length).
    // Lame lame lame!
#define MAX_TOC_LENGTH (4 + 8 * 100) // header, plus space for the maximum number of tracks on a CD.
    length = MAX_TOC_LENGTH; 
    
    while (retryCount--) {
        unsigned int newLength;
        
        OBASSERT(_generic);

        [tocData release];
        tocData = [[NSMutableData alloc] initWithLength: length];

        [response release];
        response = [_generic newResponseByReadingData: tocData
                                               opcode: 0x43 // READ TOC/PMA/ATIP
                                                 name: "READ TOC/PMA/ATIP"
                                                 arg1: 0x0  // RESERVED/RESERVED/MSF/RESERVED
                                                 arg2: 0x0  // RESERVED/FORMAT
                                                 arg3: 0x0  // RESERVED
                                                 arg4: 0x0  // RESERVED
                                                 arg5: 0x0  // RESERVED
                                                 arg6: 0x0  // Starting Track
                                                 arg7: (length >> 8) & 0xff
                                                 arg8: (length >> 0) & 0xff
                                                 arg9: 0x0   // RESERVED
                                                arg10: 0x0]; // RESERVED
        
        if ([response isError]) {
            if (![response isMediaMissingError])
                NSLog(@"Error in tableOfContents - %@", [response errorString]);
            [tocData release];
            [response release];
            return nil;
        }

        bytes = [tocData mutableBytes];

        // Figure out the actual length of the TOC data, remembering that only
        // two of the four byte header is already included in the length returned.
        newLength = MAX(4, 2 + ((bytes[0] << 8) | bytes[1]));

        if (length == MAX_TOC_LENGTH && newLength > 4) {
            length = newLength;
            // Early out for well behaved drives
            break;
        }
        length = newLength;
    }

    // Figure out how many tracks we got.
    // Eight bytes per entry, four bytes in the header.
    descriptionCount = (length - 4) / 8;
    if (!descriptionCount) {
        // Must have at least one track
        [response release];
        [tocData release];
        return nil;
    }

    descriptions = NSZoneMalloc(NULL, sizeof(*descriptions) * descriptionCount);
    bytes += 4; // skip the header
    for (descriptionIndex = 0; descriptionIndex < descriptionCount; descriptionIndex++) {
        descriptions[descriptionIndex].trackNumber = bytes[2];
        descriptions[descriptionIndex].startingLBA = (bytes[4] << 24) | (bytes[5] << 16) | (bytes[6] << 8) | bytes[7];

        bytes += 8; // next record
    }

    _tableOfContents = [[OSCSICDTableOfContents alloc] initWithTableOfContentsDescriptions: descriptions
                                                                                     count: descriptionCount];

    [response release];
    [tocData release];

    [[NSNotificationCenter defaultCenter] postNotificationName: OSCSICDAudioUnitTableOfContentsChangedNotification
                                                        object: self];

    return _tableOfContents;
}

- (void) startPlayingLBARange: (NSRange) lbaRange;
{
    OSCSIResponse             *response;

    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: nil
                                           opcode: 0xA5 // PLAY AUDIO(12)
                                             name: "PLAY AUDIO(12)"
                                             arg1: 0x0  // RESERVED/RESERVED/RELADR
                                             arg2: (lbaRange.location >> 24) & 0xff
                                             arg3: (lbaRange.location >> 16) & 0xff
                                             arg4: (lbaRange.location >>  8) & 0xff
                                             arg5: (lbaRange.location >>  0) & 0xff
                                             arg6: (lbaRange.length >> 24) & 0xff
                                             arg7: (lbaRange.length >> 16) & 0xff
                                             arg8: (lbaRange.length >>  8) & 0xff
                                             arg9: (lbaRange.length >>  0) & 0xff
                                            arg10: 0x0]; // RESERVED

    if ([response isError])
        NSLog(@"Error in startPlayingLBARange:(%d,%d) - %@",
              lbaRange.location, lbaRange.location + lbaRange.length,
              [response errorString]);

    [response release];
}

- (void) stopPlaying;
{
    [self startStopUnit: NO];
}

- (void) setAudioIsPaused: (BOOL) paused;
{
    OSCSIResponse *response;

    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: nil
                                           opcode: 0x4B      // PAUSE/RESUME
                                             name: "PAUSE/RESUME"
                                             arg1: 0x0       // RESERVED
                                             arg2: 0x0       // RESERVED
                                             arg3: 0x0       // RESERVED
                                             arg4: 0x0       // RESERVED
                                             arg5: 0x0       // RESERVED
                                             arg6: 0x0       // RESERVED
                                             arg7: 0x0       // RESERVED
                                             arg8: !paused]; // resume

    if ([response isError])
        NSLog(@"Error in setAudioIsPaused: - %@", [response errorString]);

    [response release];
}

- (OSCSILogicalBlockAddress) currentLBA;
{
    return _currentLBA;
}

- (OSCSICDAudioStatus) currentStatus;
{
    return _currentStatus;
}

//
// OSCSICDRomAudioVolumeControl
//
- (float) leftVolumePercentage;
{
    return _leftVolumePercentage;
}

- (float) rightVolumePercentage;
{
    return _rightVolumePercentage;
}

- (void) setLeftVolumePercentage: (float) leftPercentage
           rightVolumePercentage: (float) rightPercentage;
{
    unsigned char  leftVolume, rightVolume;
    OSCSIResponse *response;

    // At least some players hate it when you set the CD AUDIO CONTROL
    // page and you pass a block length of something other than zero.
    NSMutableData *modeSenseData, *modeSetData;
    struct {
        mode_sel_hdr_t  mode_header;
        mode_sel_bd_t   mode_blockdescriptor;
        unsigned char   mode_data[16];
    } *modeSenseBytes;
    struct {
        mode_sel_hdr_t  mode_header;
        unsigned char   mode_data[16];
    } *modeSetBytes;
    
    // Right now we assume that the player supports the full range of 0x0..0xff
    // for volume.  Some players may only support 2 levels (on/off) and we could
    // determine this by looking at the capabilities page, but there isn't much
    // advantage to doing so.

    if (leftPercentage < 0.0)
        leftPercentage = 0.0;
    else if (leftPercentage > 1.0)
        leftPercentage = 1.0;

    if (rightPercentage < 0.0)
        rightPercentage = 0.0;
    else if (rightPercentage > 1.0)
        rightPercentage = 1.0;


    leftVolume  = 0xFF * leftPercentage;
    rightVolume = 0xFF * rightPercentage;

    // Read the current page
    modeSenseData = [[NSMutableData alloc] initWithLength: sizeof(*modeSenseBytes)];
    OBASSERT(_generic);
    response = [_generic newResponseByReadingData: modeSenseData
                                           opcode: 0x1A  // MODE SENSE
                                             name: "MODE SENSE"
                                             arg1: 0x0   // RESERVED/DBD/RESERVED
                                             arg2: 0x0E  // PC/PAGE CODE (CD Audio Control Page)
                                             arg3: 0x0   // RESERVED
                                             arg4: sizeof(*modeSenseBytes)];

    if ([response isError]) {
        NSLog(@"Error reading mode page in -setLeftVolumePercentage:rightPercentage: - %@", [response errorString]);
        [response release];
        [modeSenseData release];
        return;
    }
    [response release];

    // Transfer the relevate parts to the modeSetData
    modeSenseBytes = [modeSenseData mutableBytes];
    modeSetData = [[NSMutableData alloc] initWithLength: sizeof(*modeSetBytes)];
    modeSetBytes = [modeSetData mutableBytes];
    modeSetBytes->mode_header = modeSenseBytes->mode_header;

    // Only used by MODE SENSE
    modeSetBytes->mode_header.msh_sd_length_0 = 0;

    // This doesn't seem to be mandated to be zero, but some players get upset if it isn't.
    modeSetBytes->mode_header.msh_med_type = 0;
    
    modeSetBytes->mode_header.msh_bd_length = 0;
    memmove(modeSetBytes->mode_data, modeSenseBytes->mode_data, sizeof(modeSetBytes->mode_data));
    [modeSenseData release];

    // Update the page
    modeSetBytes->mode_data[9]  = leftVolume;
    modeSetBytes->mode_data[11] = rightVolume;

    // Update the page
    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: modeSetData
                                           opcode: 0x15  // MODE SELECT (6)
                                             name: "MODE SELECT"
                                             arg1: 1<<4  // PF -- We're using standard page format blocks
                                             arg2: 0x0   // RESERVED
                                             arg3: 0x0   // RESERVED
                                             arg4: sizeof(*modeSetBytes)];
    if ([response isError])
        NSLog(@"Error writing mode page in -setLeftVolumePercentage:rightPercentage: - %@", [response errorString]);

    [response release];
    [modeSetData release];
}


@end


@implementation OSCSI2CDRomUnit (PrivateAPI)

- (void) _readSubcodeStatus;
{
    OSCSIResponse             *response;
    NSMutableData             *data;
    unsigned int               length;
    unsigned char             *bytes;

    if (!_mediaPresent)
        return;

    // header + subcode size for this format
    length = 4 + 12;


    data = [[NSMutableData alloc] initWithLength: length];
    OBASSERT(_generic);
    response = [_generic newResponseByReadingData: data
                                           opcode: 0x42 // READ SUB-CHANNEL
                                             name: "READ SUB-CHANNEL"
                                             arg1: 0x0  // RESERVED/MSF/RESERVED (we wan't LBA)
                                             arg2: 1<<6 // RESERVED/SUBQ/RESERVED
                                             arg3: 0x1  // CD Current position data format
                                             arg4: 0x0  // RESERVED
                                             arg5: 0x0  // RESERVED
                                             arg6: 0x0  // Track Number not used for this format
                                             arg7: (length >> 8) & 0xff
                                             arg8: (length >> 0) & 0xff];

    if ([response isError])
        NSLog(@"Error in _readSubcodeStatus - %@", [response errorString]);

    bytes = [data mutableBytes];

    // This byte will also tell use whether we are playing or not,
    // but we can get that from the MECHANISM STATUS command too
    // (plus whether we are scanning).
    switch (bytes[1]) {
        case 0x11:
            _currentStatus = OSCSICDAudioPlayingStatus;
            break;
        case 0x12:
            _currentStatus = OSCSICDAudioPausedStatus;
            break;
        case 0x13:
        case 0x14:
        case 0x15:
            _currentStatus = OSCSICDAudioIdleStatus;
        default:
            _currentStatus = OSCSICDAudioUnknownStatus;
            break;
    }



    bytes += 4; // skip the header

    _currentTrack = bytes[2];
    _currentIndex = bytes[3];

    // MECHANISM STATUS doesn't seem to return the current LBA
    // on all drives, even though the standard says it should.
    // Reading the subcode channel is much more reliable.  Also,
    // MECHANISM STATUS is only supported on SCSI-3 compliant drives.
    _currentLBA   = (bytes[4] << 24) | (bytes[5] << 16) | (bytes[6] << 8) | bytes[7];
}

- (void) _readControlPage;
{
    OSCSIResponse *response;
    NSMutableData *modeData;
    struct {
        mode_sel_hdr_t  mode_header;
        mode_sel_bd_t   mode_blockdescriptor;
        unsigned char   mode_data[16];
    } *modeBytes;
    
    modeData = [[NSMutableData alloc] initWithLength: sizeof(*modeBytes)];
    OBASSERT(_generic);
    response = [_generic newResponseByReadingData: modeData
                                           opcode: 0x1A  // MODE SENSE
                                             name: "MODE SENSE"
                                             arg1: 0x0   // RESERVED/DBD/RESERVED
                                             arg2: 0x0E  // PC/PAGE CODE (CD Audio Control Page)
                                             arg3: 0x0   // RESERVED
                                             arg4: sizeof(*modeBytes)];

    if ([response isError])
        NSLog(@"Error reading mode page in _readControlPage - %@", [response errorString]);

    modeBytes = [modeData mutableBytes];

    // Update the page
    _leftVolumePercentage  = (float)modeBytes->mode_data[9]  / (float)0xff;
    _rightVolumePercentage = (float)modeBytes->mode_data[11] / (float)0xff;

    [response release];
    [modeData release];
}


@end
