// Copyright 1998-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniSCSI/OSCSI3CDRomUnit+CDRecorder.h>

#import <OmniSCSI/OSCSIGeneric.h>
#import <OmniSCSI/OSCSIResponse.h>
#import <OmniSCSI/OSCSICDTableOfContents.h>

#import <dev/scsireg.h>

@implementation OSCSI3CDRomUnit (CDRecorderSupport)

- (unsigned short) maximumWriteSpeed;
{
    OBPRECONDITION(_isWritable);

    return _maximumWriteSpeed;
}

- (void) beginWritingWithTestMode: (BOOL) testMode;
{
    NSMutableData *modeSenseData, *modeSelectData;
    const struct {
        mode_sel_hdr_t  mode_header;
        mode_sel_bd_t   mode_blockdescriptor;
        unsigned char   mode_data[55];
    } *modeSenseBytes;
    struct {
        mode_sel_hdr_t  mode_header;
        mode_sel_bd_t   mode_blockdescriptor;
        unsigned char   mode_data[55];
    } *modeSelectBytes;
    OSCSIResponse *response;
    
    OBPRECONDITION(_isWritable);

    modeSenseData = [[NSMutableData alloc] initWithLength: sizeof(*modeSenseBytes)];
    OBASSERT(_generic);
    response = [_generic newResponseByReadingData: modeSenseData
                                           opcode: 0x1A  // MODE SENSE
                                             name: "MODE SENSE"
                                             arg1: 0x0   // RESERVED/DBD/RESERVED
                                             arg2: 0x05  // PC/PAGE CODE (Write Paramters Mode Page)
                                             arg3: 0x0   // RESERVED
                                             arg4: sizeof(*modeSenseBytes)];
    if ([response isError]) {
        NSLog(@"Error reading Write Paramters Mode Page in %@ - %@", NSStringFromSelector(_cmd), [response errorString]);
        [response release];
        [modeSenseData release];
        return;
    }
    [response release];

    // Transfer the relevate parts to the modeSelectData
    modeSenseBytes = [modeSenseData bytes];
    modeSelectData = [[NSMutableData alloc] initWithLength: sizeof(*modeSelectBytes)];
    modeSelectBytes = [modeSelectData mutableBytes];
    
    modeSelectBytes->mode_header = modeSenseBytes->mode_header;
    modeSelectBytes->mode_header.msh_sd_length_0 = 0;    // Only used by MODE SENSE
    modeSelectBytes->mode_header.msh_med_type = 0; // This doesn't seem to be mandated to be zero, but some players get upset if it isn't.

    modeSelectBytes->mode_blockdescriptor = modeSenseBytes->mode_blockdescriptor;
#ifdef __BIG_ENDIAN__
    modeSelectBytes->mode_blockdescriptor.msbd_blocklength = OSCSI_CDDA_BYTES_PER_BLOCK;
#else
    modeSelectBytes->mode_blockdescriptor.msbd_blocklength2 = (OSCSI_CDDA_BYTES_PER_BLOCK & 0xff0000) >> 16;
    modeSelectBytes->mode_blockdescriptor.msbd_blocklength1 = (OSCSI_CDDA_BYTES_PER_BLOCK & 0x00ff00) >>  8;
    modeSelectBytes->mode_blockdescriptor.msbd_blocklength0 = (OSCSI_CDDA_BYTES_PER_BLOCK & 0x0000ff) >>  0;
#endif
    
    memmove(modeSelectBytes->mode_data, modeSenseBytes->mode_data, sizeof(modeSelectBytes->mode_data));
    [modeSenseData release];

    // Change the stuff we want
    modeSelectBytes->mode_data[2] = ((testMode ? 1 : 0) << 4) | (0x2);  // testMode/session-at-once
    modeSelectBytes->mode_data[3] = 0x2;  // next session disabled, not a copy, 2 channel, no preemphasis, digital copy permitted
    modeSelectBytes->mode_data[4] = 0x0;  // raw data
    modeSelectBytes->mode_data[8] = 0x0;  // CDDA/CDROM session
    modeSelectBytes->mode_data[14] = 0x00;  // audio pause length MSB
    modeSelectBytes->mode_data[15] = 0x96;  // audio pause length LSB (150 frames == 2 seconds)


    modeSelectBytes->mode_data[16] = 0x0;  // MCN not valid
    modeSelectBytes->mode_data[32] = 0x0;  // ISRC not valid

    // Change the page!
    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: modeSelectData
                                           opcode: 0x15  // MODE SELECT (6)
                                             name: "MODE SELECT"
                                             arg1: 1<<4  // PF -- We're using standard page format blocks
                                             arg2: 0x0   // RESERVED
                                             arg3: 0x0   // RESERVED
                                             arg4: sizeof(*modeSelectBytes)];
    if ([response isError])
        NSLog(@"Error writing mode page in %@ - %@", NSStringFromSelector(_cmd), [response errorString]);

    [response release];
    [modeSelectData release];

#ifdef DEBUG
    // Read back the page
    modeSenseData = [[NSMutableData alloc] initWithLength: sizeof(*modeSenseBytes)];
    OBASSERT(_generic);
    response = [_generic newResponseByReadingData: modeSenseData
                                           opcode: 0x1A  // MODE SENSE
                                             name: "MODE SENSE"
                                             arg1: 0x0   // RESERVED/DBD/RESERVED
                                             arg2: 0x05  // PC/PAGE CODE (Write Paramters Mode Page)
                                             arg3: 0x0   // RESERVED
                                             arg4: sizeof(*modeSenseBytes)];
    [modeSenseData release];
    [response release];
#endif
}

- (void) setWriteSpeed: (unsigned short) writeKBPerSecond;
{
    OSCSIResponse *response;

    if (!writeKBPerSecond)
        // This will set the speed to the maximum supported speed
        writeKBPerSecond = 0xffff;
    
    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: nil
                                           opcode: 0xBB // SET CD SPEED
                                             name: "SET CD SPEED"
                                             arg1: 0x0  // RESERVED
                                             arg2: (_currentReadSpeed & 0xff00) >> 8
                                             arg3: (_currentReadSpeed & 0x00ff) >> 0
                                             arg4: (writeKBPerSecond & 0xff00) >> 8
                                             arg5: (writeKBPerSecond & 0x00ff) >> 0
                                             arg6: 0x0  // RESERVED
                                             arg7: 0x0  // RESERVED
                                             arg8: 0x0  // RESERVED
                                             arg9: 0x0  // RESERVED
                                            arg10: 0x0]; // RESERVED

    if ([response isError])
        NSLog(@"Error setting write speed - %@", [response errorString]);
    [response release];
}

typedef struct _OSCSICDSendTOCEntry {
    unsigned char ctlAdr;
    unsigned char tno;
    unsigned char index;
    unsigned char dataForm;
    unsigned char scms;
    unsigned char minutes, seconds, frames;
} OSCSICDSendTOCEntry;

- (void) setTableOfContents: (OSCSICDTableOfContents *) toc;
{
    OSCSIResponse            *response;
    unsigned int              tocLength;
    NSMutableData            *tocData;
    OSCSICDSendTOCEntry      *tocEntries;
    unsigned int              trackNumber, trackCount;
    OSCSIMinuteSecondsFrames  msf;

    trackCount = [toc numberOfTracks];

    // For some reason the CTL/ADR field is swapped between the TOC you get
    // when you read it off an existing CD and the TOC you write here.
    // Go figure.

    // Allocate enough space for all the tracks plus the lead-in and lead-out.
    // Later, we'll need to support ISRCs and catalogue codes.
    tocLength  = sizeof(*tocEntries) * (trackCount + 3);
    tocData    = [[NSMutableData alloc] initWithLength: tocLength];
    tocEntries = [tocData mutableBytes];

    // Lead-in
    tocEntries->ctlAdr   = 0x1;
    tocEntries->tno      = 0;
    tocEntries->index    = 0;
    tocEntries->dataForm = 0x0;
    tocEntries->scms     = 0;
    tocEntries->minutes  = 0;
    tocEntries->seconds  = 0;
    tocEntries->frames   = 0;
    tocEntries++;

    tocEntries->ctlAdr   = 0x1;
    tocEntries->tno      = 1;
    tocEntries->index    = 0;
    tocEntries->dataForm = 0x0;
    tocEntries->scms     = 0;
    tocEntries->minutes  = 0;
    tocEntries->seconds  = 0;
    tocEntries->frames   = 0;
    tocEntries++;
    
    // Program area
    for (trackNumber = 0; trackNumber < trackCount; trackNumber++) {
        // tracks are 1-based indexing
        msf = OSCSIAbsoluteLogicalBlockAddressToMinutesSecondsFrames([toc startingLBAForTrackNumber: trackNumber + 1]);
        
        tocEntries->ctlAdr   = (0x0 /*2*/ << 4) | (0x1);  // 2 audio, no pre-emph, allow digital copy, start time
        tocEntries->tno      = trackNumber + 1; // tracks are 1-based indexing
        tocEntries->index    = 1; // we don't support pre-gap yet
        tocEntries->dataForm = 0x0;  // subcode holds current time, data is CDDA
        tocEntries->scms     = 0;
        tocEntries->minutes  = msf.minutes;
        tocEntries->seconds  = msf.seconds;
        tocEntries->frames   = msf.frames;
        tocEntries++;
    }

    // Lead-out
    msf = OSCSIAbsoluteLogicalBlockAddressToMinutesSecondsFrames([toc startingLBAForTrackNumber: OSCSI_LEAD_OUT_TRACK]);
    tocEntries->ctlAdr   = 0x1;
    tocEntries->tno      = OSCSI_LEAD_OUT_TRACK;
    tocEntries->index    = 1;
    tocEntries->dataForm = 0x0;
    tocEntries->scms     = 0;
    tocEntries->minutes  = msf.minutes;
    tocEntries->seconds  = msf.seconds;
    tocEntries->frames   = msf.frames;

    // Write it!
    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: tocData
                                           opcode: 0x5D // SEND CUE SHEET
                                             name: "SEND CUE SHEET"
                                             arg1: 0x0  // RESERVED
                                             arg2: 0x0  // RESERVED
                                             arg3: 0x0  // RESERVED
                                             arg4: 0x0  // RESERVED
                                             arg5: 0x0  // RESERVED
                                             arg6: (tocLength & 0xff0000) >> 16
                                             arg7: (tocLength & 0x00ff00) >>  8
                                             arg8: (tocLength & 0x0000ff) >>  0];

    if ([response isError])
        NSLog(@"Error sending cue sheet - %@", [response errorString]);
    [tocData release];
    [response release];
}

- (void) writeData: (NSData *) data atAddress: (OSCSILogicalBlockAddress) lba;
{
    OSCSIResponse            *response;
    unsigned long             dataLength;
    OSCSILogicalBlockAddress  blockCount;
    
    OBPRECONDITION(_isWritable);

    dataLength = [data length];
    blockCount = dataLength / OSCSI_CDDA_BYTES_PER_BLOCK;
    if (dataLength % OSCSI_CDDA_BYTES_PER_BLOCK) {
        NSLog(@"-[%@ %@]: data is not an even multiple of %d.  Padding with zeros.",
              isa, NSStringFromSelector(_cmd), OSCSI_CDDA_BYTES_PER_BLOCK);
        blockCount++;
        data = [[data mutableCopy] autorelease];
        [(NSMutableData *)data setLength: blockCount * OSCSI_CDDA_BYTES_PER_BLOCK];
    }

    if (blockCount > 0xffff)
        [NSException raise: NSInvalidArgumentException
                    format: @"Attempted to write more thatn %d blocks in a single invocation of -%@.",
            0xffff, NSStringFromSelector(_cmd)];

    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: data
                                           opcode: 0x2A // WRITE
                                             name: "WRITE"
                                             arg1: 0x0  // DPO/FUA/RELADDR
                                             arg2: (lba & 0xff000000) >> 24
                                             arg3: (lba & 0x00ff0000) >> 16
                                             arg4: (lba & 0x0000ff00) >>  8
                                             arg5: (lba & 0x000000ff) >>  0
                                             arg6: 0x0 // RESERVED
                                             arg7: (blockCount & 0xff00) >> 8
                                             arg8: (blockCount & 0x00ff) >> 0];
    if ([response isError])
        NSLog(@"Error writing CDDA data - %@", [response errorString]);
    [response release];
}

- (void) finishWriting;
{
    OSCSIResponse            *response;

    OBPRECONDITION(_isWritable);

    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: nil
                                           opcode: 0x5B // CLOSE TRACK/SESSION
                                             name: "CLOSE TRACK/SESSION"
                                             arg1: 0x1  // Immed
                                             arg2: 0x2  // Close session, not track
                                             arg3: 0x0  // RESERVED
                                             arg4: 0x0  // RESERVED
                                             arg5: 0x0  // Track Number -- should be ignored if we are closing the session
                                             arg6: 0x0  // RESERVED
                                             arg7: 0x0  // RESERVED
                                             arg8: 0x0]; // RESERVED
    if ([response isError])
        NSLog(@"Error closing CDDA session - %@", [response errorString]);
    [response release];
}

@end
