// Copyright 1998-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniSCSI/OSCSI3CDRomUnit.h>
#import <OmniSCSI/OSCSI3CDRomUnit+CDRecorder.h>

#import <OmniSCSI/OSCSIResponse.h>
#import <OmniSCSI/OSCSIGeneric.h>
#import <OmniSCSI/OSCSICDTableOfContents.h>

#import <OmniBase/OmniBase.h>
#import <dev/scsireg.h>

NSString *OSCSICDAudioUnitTableOfContentsChangedNotification = @"OSCSICDAudioUnitTableOfContentsChangedNotification";

@interface OSCSI3CDRomUnit (PrivateAPI)
- (void) _readMechanismStatus;
@end

@implementation OSCSI3CDRomUnit

+ (NSString *) displayName;
{
    return @"SCSI-3 Multimedia Unit";
}

- initWithTarget: (OSCSITarget *) target
      unitNumber: (OSCSIUnitNumberType) unitNumber
 unitDescription: (OSCSIUnitDescription *) unitDescription
temporaryGeneric: (OSCSIGeneric *) temporaryGeneric;
{
    const struct {
        mode_sel_hdr_t mode_header;
        mode_sel_bd_t  mode_blockdescriptor;
        unsigned char  mode_data[21];
    } *modeSenseBytes;
    NSMutableData *modeSenseData;
    OSCSIResponse *response;

    if (!(self = [super initWithTarget: target
                            unitNumber: unitNumber
                       unitDescription: unitDescription
                      temporaryGeneric: temporaryGeneric]))
        return nil;

    // Read the CD Capabilities page.
    modeSenseData = [[NSMutableData alloc] initWithLength: sizeof(*modeSenseBytes)];
    OBASSERT(temporaryGeneric);
    response = [temporaryGeneric newResponseByReadingData: modeSenseData
                                                   opcode: 0x1A // MODE SENSE
                                                     name: "MODE SENSE"
                                                     arg1: 0x0   // RESERVED/DBD/RESERVED
                                                     arg2: 0x2A  // PC/PAGE CODE (CD Capabilities and Mechanical Status Page)
                                                     arg3: 0x0   // RESERVED
                                                     arg4: sizeof(*modeSenseBytes)];
    if ([response isError]) {
        NSLog(@"Error reading CD Capabilities and Mechanical Status Page in %@ - %@",
              NSStringFromSelector(_cmd), [response errorString]);
        [modeSenseData release];
        [response release];
        [self release];
        return nil;
    }

    modeSenseBytes = [modeSenseData bytes];

    _isWritable            = (modeSenseBytes->mode_data[3] & (0x1 << 0)) != 0;
    _isReWritable          = (modeSenseBytes->mode_data[3] & (0x1 << 1)) != 0;
    _supportsTestWriteMode = (modeSenseBytes->mode_data[3] & (0x1 << 2)) != 0;
    _cddaSupported         = (modeSenseBytes->mode_data[5] & (0x1 << 0)) != 0;
    _cddaStreamAccurate    = (modeSenseBytes->mode_data[5] & (0x1 << 1)) != 0;

    _maximumReadSpeed  = (modeSenseBytes->mode_data[ 8] << 8) | modeSenseBytes->mode_data[ 9];
    _currentReadSpeed  = (modeSenseBytes->mode_data[14] << 8) | modeSenseBytes->mode_data[15];
    _maximumWriteSpeed = (modeSenseBytes->mode_data[18] << 8) | modeSenseBytes->mode_data[19];
    _currentWriteSpeed = (modeSenseBytes->mode_data[20] << 8) | modeSenseBytes->mode_data[21];

    [response release];
    [modeSenseData release];

    return self;
}

- (id <OSCSICDRecorder>) recorderUnit;
{
    if (_isWritable)
        return self;
    else
        return nil;
}

//
// OSCSICDRomAudioControl overrides
//

- (void) updateStatus;
{
    [super updateStatus];
    [self _readMechanismStatus];
}

- (void) stopPlaying;
{
    // The SCSI-2 superclass uses START/STOP UNIT.  This will probably work on all
    // players, but this is actually in the SCSI-3 spec and it seems more likely
    // to stop scanning too.
    OSCSIResponse             *response;

    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: nil
                                           opcode: 0x4E // STOP PLAY/SCAN
                                             name: "STOP PLAY/SCAN"
                                             arg1: 0x0  // RESERVED
                                             arg2: 0x0  // RESERVED
                                             arg3: 0x0  // RESERVED
                                             arg4: 0x0  // RESERVED
                                             arg5: 0x0  // RESERVED
                                             arg6: 0x0  // RESERVED
                                             arg7: 0x0  // RESERVED
                                             arg8: 0x0  // RESERVED
                                             arg9: 0x0  // RESERVED
                                            arg10: 0x0]; // RESERVED

    if ([response isError])
        NSLog(@"Error in stopPlaying - %@", [response errorString]);

    [response release];
}

//
// OSCSICDRomAudioIdentification
//
#warning When this is implemented it might actually be able to be moved into the SCSI2 superclass
- (NSString *) mediaCatalogueNumber;
{
    OBRequestConcreteImplementation(self, _cmd);
    return @"";
}

- (NSString *) internationalStandardRecordingCodeForTrack: (unsigned char) trackNumber;
{
    OBRequestConcreteImplementation(self, _cmd);
    return @"";
}

//
// OSCSICDRomAudioScan
//
- (void) startScanningFromLBA: (OSCSILogicalBlockAddress) lba
                      forward: (BOOL) forward;
{
    OSCSIResponse *response;

    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: nil
                                           opcode: 0xBA  // SCAN
                                             name: "SCAN"
                                             arg1: (!forward ? 1 : 0) << 4 // RESERVED/DIRECT/RESERVED/RELADR
                                             arg2: (lba >> 24) & 0xff
                                             arg3: (lba >> 16) & 0xff
                                             arg4: (lba >>  8) & 0xff
                                             arg5: (lba >>  0) & 0xff
                                             arg6: 0x0   // RESERVED
                                             arg7: 0x0   // RESERVED
                                             arg8: 0x0   // RESERVED
                                             arg9: 0x0   // TYPE (LBA)/RESERVED
                                            arg10: 0x0]; // RESERVED

    if ([response isError])
        NSLog(@"Error writing mode page in startScanningFromLBA:forward: - %@", [response errorString]);

    [response release];
}

- (void) stopScanningAndResumePlayingUntilLBA: (OSCSILogicalBlockAddress) lba;
{
    // Stopping the SCAN uses the same command as stopping the PLAY
    [self stopPlaying];

    [self updateStatus];
    if (_currentLBA < lba)
        [self startPlayingLBARange: NSMakeRange(OSCSI_USE_CURRENT_PLAY_POSITION, lba - _currentLBA)];
}

//
// OSCSICDRomAudioData
//

- (NSData *) readAudioDataAtLBA: (OSCSILogicalBlockAddress) lba
                     blockCount: (unsigned int) blockCount;
{
    OSCSIResponse *response;
    NSMutableData *data;
    unsigned int   byteCount;

    byteCount = blockCount * OSCSI_CDDA_BYTES_PER_BLOCK;
    data = [NSMutableData dataWithLength: byteCount];

    OBASSERT(_generic);
    response = [_generic newResponseByReadingData: data
                                           opcode: 0xBE // READ CD
                                             name: "READ CD"
                                             arg1: 0x0  // RESERVED/Sector Type/RESERVED/RELADR -- Guessing on the Sector Type
                                             arg2: (lba & 0xff000000) >> 24
                                             arg3: (lba & 0x00ff0000) >> 16
                                             arg4: (lba & 0x0000ff00) >>  8
                                             arg5: (lba & 0x000000ff) >>  0
                                             arg6: (blockCount & 0xff0000) >> 16
                                             arg7: (blockCount & 0x00ff00) >>  8
                                             arg8: (blockCount & 0x0000ff) >>  0
                                             arg9: 0x10  // CDDA
                                            arg10: 0x0]; // RESERVED/Sub-channel selection bits

    if ([response isError]) {
        NSLog(@"Error reaing CDDA data in -readAudioDataAtLBA:blockCount: - %@", [response errorString]);
        data = nil;
    } else {
        int bytesNotTransfered;
        
        // If we only read part of the requested data, truncate our data object
        if ((bytesNotTransfered = [response bytesNotTransfered]))
            [data setLength: byteCount - bytesNotTransfered];
        
#ifdef __BIG_ENDIAN__
        {
            OSCSICDAudioSample *samples;
            unsigned int        sampleCount;

            // CDDA data is in little-endian format on the disc.  Swap it.
            samples = [data mutableBytes];
            sampleCount = OSCSI_CDDA_SAMPLES_PER_BLOCK * blockCount;
            while (sampleCount--) {
                samples->left  = NSSwapLittleShortToHost(samples->left);
                samples->right = NSSwapLittleShortToHost(samples->right);
                samples++;
            }
        }
#endif
    }

    [response release];
    return data;
}

@end

@implementation OSCSI3CDRomUnit (PrivateAPI)


- (void) _readMechanismStatus;
{
    OSCSIResponse *response;
    NSMutableData *data;
    unsigned char *bytes;
    unsigned int   length;

    if (!_mediaPresent)
        return;
    
    // The header is 8 bytes, each slot (max of 32) is 4 bytes
#warning We should check the CD capabilities page to see if we are a changer
    //length = 8 + 4 * 32;
    length = 8;
    
    data = [[NSMutableData alloc] initWithLength: length];

    OBASSERT(_generic);
    response = [_generic newResponseByReadingData: data
                                           opcode: 0xBD // MECHANISM STATUS
                                             name: "MECHANISM STATUS"
                                             arg1: 0x0  // RESERVED
                                             arg2: 0x0  // RESERVED
                                             arg3: 0x0  // RESERVED
                                             arg4: 0x0  // RESERVED
                                             arg5: 0x0  // RESERVED
                                             arg6: 0x0  // RESERVED
                                             arg7: 0x0  // RESERVED
                                             arg8: (length >> 8) & 0xff
                                             arg9: (length >> 0) & 0xff
                                            arg10: 0x0]; // RESERVED

    if ([response isError]) {
        NSLog(@"Error in _readMechanismStatus - %@", [response errorString]);
    } else {
        
        bytes = [data mutableBytes];

        // Right now, the only thing this gives us that we can't get from the _readSubcodeStatus
        // method is whether we are scanning or not.
        if (((bytes[1] >> 5) & 0x7) == 0x2)
            _currentStatus = OSCSICDAudioScanningStatus;
    }

    [response release];
    [data release];
}

@end

