// Copyright 1998-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniSCSI/OSCSIATAPICDRomUnit.h>

#import <OmniSCSI/OSCSIGeneric.h>
#import <OmniSCSI/OSCSIResponse.h>
#import <OmniSCSI/OSCSICDProtocols.h>
#import <OmniSCSI/OSCSICDTableOfContents.h>

@implementation OSCSIATAPICDRomUnit

// ATAPI has doesn't support the command that the SCSI2 class uses
- (void) startPlayingLBARange: (NSRange) lbaRange;
{
    OSCSIResponse             *response;
    OSCSIMinuteSecondsFrames   startMSF, endMSF;
    OSCSILogicalBlockAddress   startLBA, endLBA, leadOut;

    startLBA = lbaRange.location;
    endLBA = lbaRange.location + lbaRange.length;

    // At least some ATAPI drives get ill if you try to play to the first frame of the
    // lead out track.  We'll clip the range so that it doesn't specify any frame
    // at or beyond the leadout LBA.
    // to attempt to play any closer than N to the end of the disc.
    leadOut = [_tableOfContents startingLBAForTrackNumber: OSCSI_LEAD_OUT_TRACK];
    if (endLBA >= leadOut)
        endLBA = leadOut - 1;

    startMSF = OSCSIAbsoluteLogicalBlockAddressToMinutesSecondsFrames(startLBA);
    endMSF = OSCSIAbsoluteLogicalBlockAddressToMinutesSecondsFrames(endLBA);
    
        
    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: nil
                                           opcode: 0x47 // PLAY AUDIO MSF
                                             name: "PLAY AUDIO MSF"
                                             arg1: 0x0 // RESERVED
                                             arg2: 0x0 // RESERVED
                                             arg3: startMSF.minutes
                                             arg4: startMSF.seconds
                                             arg5: startMSF.frames
                                             arg6: endMSF.minutes
                                             arg7: endMSF.seconds
                                             arg8: endMSF.frames
                                             arg9: 0x0   // RESERVED
                                            arg10: 0x0]; // RESERVED

    if ([response isError])
        NSLog(@"Error in startPlayingLBARange:(%d,%d) - %@",
              lbaRange.location, lbaRange.location + lbaRange.length,
              [response errorString]);

    [response release];
}

// These are private methods on our superclass -- in the ATAPI world, we need to
// use the 10-byte MODE SENSE/SELECT commands.  For some reason or other, I couldn't
// get the 10-byte commands to work on some SCSI-2 drives (although the 6-byte
// version works great).

- (void) setLeftVolumePercentage: (float) leftPercentage
           rightVolumePercentage: (float) rightPercentage;
{
    unsigned char  leftVolume, rightVolume;
    OSCSIResponse *response;

    NSMutableData *modeSenseData;
    struct {
        unsigned char  modeParameterHeader[8];
        unsigned char  mode_data[16];
    } *modeSenseBytes;

    // Right now we assume that the player supports the full range of 0x0..0xff
    // for volume.  Some players may only support 2 levels (on/off) and we could
    // determine this by looking at the capabilities page, but there isn't much
    // advantage to doing so.

    if (leftPercentage < 0.0)
        leftPercentage = 0.0;
    else if (leftPercentage > 1.0)
        leftPercentage = 1.0;

    if (rightPercentage < 0.0)
        rightPercentage = 0.0;
    else if (rightPercentage > 1.0)
        rightPercentage = 1.0;


    leftVolume  = 0xFF * leftPercentage;
    rightVolume = 0xFF * rightPercentage;

    // Read the current page.  Use the 10-byte command instead of the 6-byte since
    // that is the one in the ATAPI spec as well.
    modeSenseData = [[NSMutableData alloc] initWithLength: sizeof(*modeSenseBytes)];
    OBASSERT(_generic);
    response = [_generic newResponseByReadingData: modeSenseData
                                           opcode: 0x5A  // MODE SENSE (10)
                                             name: "MODE SENSE"
                                             arg1: 0x0   // RESERVED/DBD/RESERVED
                                             arg2: 0x0E  // PC/PAGE CODE (CD Audio Control Page)
                                             arg3: 0x0   // RESERVED
                                             arg4: 0x0   // RESERVED
                                             arg5: 0x0   // RESERVED
                                             arg6: 0x0   // RESERVED
                                             arg7: (sizeof(*modeSenseBytes) & 0xff00) >> 8
                                             arg8: (sizeof(*modeSenseBytes) & 0x00ff) >> 0];

    if ([response isError]) {
        NSLog(@"Error reading mode page in -setLeftVolumePercentage:rightPercentage: - %@", [response errorString]);
        [response release];
        [modeSenseData release];
        return;
    }
    [response release];

    // Update the modeSenseBytes as necessary
    modeSenseBytes = [modeSenseData mutableBytes];

    // Only used by MODE SENSE
    modeSenseBytes->modeParameterHeader[0] = 0;
    modeSenseBytes->modeParameterHeader[1] = 0;

    // The medium type doesn't seem to be mandated to be zero, but some players get upset if it isn't.
    modeSenseBytes->modeParameterHeader[2] = 0;

    // Update the volume
    modeSenseBytes->mode_data[9]  = leftVolume;
    modeSenseBytes->mode_data[11] = rightVolume;

    // Update the page
    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: modeSenseData
                                           opcode: 0x55  // MODE SELECT (10)
                                             name: "MODE SELECT (10)"
                                             arg1: 1<<4  // PF -- We're using standard page format blocks
                                             arg2: 0x0   // RESERVED
                                             arg3: 0x0   // RESERVED
                                             arg4: 0x0   // RESERVED
                                             arg5: 0x0   // RESERVED
                                             arg6: 0x0   // RESERVED
                                             arg7: (sizeof(*modeSenseBytes) & 0xff00) >> 8
                                             arg8: (sizeof(*modeSenseBytes) & 0x00ff) >> 0];
    
    if ([response isError])
        NSLog(@"Error writing mode page in -setLeftVolumePercentage:rightPercentage: - %@", [response errorString]);

    [response release];
    [modeSenseData release];
}

//
// OSCSICDRomAudioData.  Exactly the same code for this command is in OSCSIATAPICDRomUnit
// and OSCSI3CDRomUnit.  It would be nice to put this somewhere, but right now OSCSI2CDRomUnit
// is the parent class and it doesn't support this command.

- (NSData *) readAudioDataAtLBA: (OSCSILogicalBlockAddress) lba
                     blockCount: (unsigned int) blockCount;
{
    OSCSIResponse *response;
    NSMutableData *data;
    unsigned int   byteCount;

    byteCount = blockCount * OSCSI_CDDA_BYTES_PER_BLOCK;
    data = [NSMutableData dataWithLength: byteCount];

    OBASSERT(_generic);
    response = [_generic newResponseByReadingData: data
                                           opcode: 0xBE // READ CD
                                             name: "READ CD"
                                             arg1: 0x0  // RESERVED/Sector Type/RESERVED/RELADR -- Guessing on the Sector Type
                                             arg2: (lba & 0xff000000) >> 24
                                             arg3: (lba & 0x00ff0000) >> 16
                                             arg4: (lba & 0x0000ff00) >>  8
                                             arg5: (lba & 0x000000ff) >>  0
                                             arg6: (blockCount & 0xff0000) >> 16
                                             arg7: (blockCount & 0x00ff00) >>  8
                                             arg8: (blockCount & 0x0000ff) >>  0
                                             arg9: 0x10  // CDDA
                                            arg10: 0x0]; // RESERVED/Sub-channel selection bits

    if ([response isError]) {
        NSLog(@"Error reaing CDDA data in -readAudioDataAtLBA:blockCount: - %@", [response errorString]);
        data = nil;
    } else {
        int bytesNotTransfered;

        // If we only read part of the requested data, truncate our data object
        if ((bytesNotTransfered = [response bytesNotTransfered]))
            [data setLength: byteCount - bytesNotTransfered];

#ifdef __BIG_ENDIAN__
        {
            OSCSICDAudioSample *samples;
            unsigned int        sampleCount;

            // CDDA data is in little-endian format on the disc.  Swap it.
            samples = [data mutableBytes];
            sampleCount = OSCSI_CDDA_SAMPLES_PER_BLOCK * blockCount;
            while (sampleCount--) {
                samples->left  = NSSwapLittleShortToHost(samples->left);
                samples->right = NSSwapLittleShortToHost(samples->right);
                samples++;
            }
        }
#endif
    }

    [response release];
    return data;
}

@end


@implementation OSCSIATAPICDRomUnit (PrivateAPI)

// We're overriding a private method.  Bad, bad, bad
- (void) _readControlPage;
{
    OSCSIResponse *response;
    NSMutableData *modeData;
    struct {
        unsigned char  mode_header[8];
        unsigned char  mode_data[16];
    } *modeBytes;

    // Read the sense page.  Use the 10-byte command that ATAPI drives like.
    modeData = [[NSMutableData alloc] initWithLength: sizeof(*modeBytes)];
    OBASSERT(_generic);
    response = [_generic newResponseByReadingData: modeData
                                           opcode: 0x5A  // MODE SENSE (10)
                                             name: "MODE SENSE"
                                             arg1: 0x0   // RESERVED/DBD/RESERVED
                                             arg2: 0x0E  // PC/PAGE CODE (CD Audio Control Page)
                                             arg3: 0x0   // RESERVED
                                             arg4: 0x0   // RESERVED
                                             arg5: 0x0   // RESERVED
                                             arg6: 0x0   // RESERVED
                                             arg7: (sizeof(*modeBytes) & 0xff00) >> 8
                                             arg8: (sizeof(*modeBytes) & 0x00ff) >> 0];

    if ([response isError])
        NSLog(@"Error reading mode page in _readControlPage - %@", [response errorString]);

    modeBytes = [modeData mutableBytes];

    // Grab the data we want
    _leftVolumePercentage  = (float)modeBytes->mode_data[9]  / (float)0xff;
    _rightVolumePercentage = (float)modeBytes->mode_data[11] / (float)0xff;

    [response release];
    [modeData release];
}

@end
