// Copyright 1998-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniSCSI/OSCSIBus.h>

#import <OmniSCSI/OSCSITarget.h>
#import <OmniSCSI/OSCSIGeneric.h>

#import <OmniBase/OmniBase.h>
#import <OmniBase/system.h>
#import <OmniBase/macros.h>


@interface OSCSIBus (PrivateAPI)
- _initWithBusNumber: (OSCSIBusNumberType) busNumber;
@end

@implementation OSCSIBus

+ (NSArray *) busses;
{
    static NSMutableArray *_busses = nil;
    OSCSIBusNumberType     busIndex, busCount;
    
    if (_busses)
        return _busses;

    _busses = [[NSMutableArray alloc] init];
    busCount = [OSCSIGeneric numberOfBusses];
    for (busIndex = 0; busIndex < busCount; busIndex++) {
        OSCSIBus *bus;

        bus = [[OSCSIBus alloc] _initWithBusNumber: busIndex];
        [_busses addObject: bus];
        [bus release];
    }

    return _busses;
}

+ (OSCSIBus *) busWithBusNumber: (OSCSIBusNumberType) busNumber;
{
    return [[self busses] objectAtIndex: busNumber];
}

- (void) dealloc;
{
    [_targets release];
    [super dealloc];
}

- (unsigned int) busNumber;
{
    return _busNumber;
}

- (NSArray *) targets;
{
    if (!_targets) {
        OSCSITargetNumberType   targetNumber;
        OSCSITarget            *scsiTarget;
        unsigned int            maximumTargetsPerBus;
        unsigned int            diskTargets = 0; // The current 'sd' index
        unsigned int            tapeTargets = 0; // The current 'st' index
        NSString               *rawDeviceName;
        
        _targets = [[NSMutableArray alloc] init];
        
        // Also, I don't see any command in the SCSI-3 SCC spec about inquiring about
        // the available devices on the SCSI-bus.  We'll just try the available combinations.
        maximumTargetsPerBus = [OSCSIGeneric maximumTargetsPerBus];

        //
        // Search for SCSI devices
        //
        for (targetNumber = 0; targetNumber < maximumTargetsPerBus; targetNumber++) {
            NS_DURING {
                scsiTarget = [[OSCSITarget alloc] initWithBusNumber: _busNumber
                                                       targetNumber: targetNumber];
                if (scsiTarget) {
                    [(NSMutableArray *)_targets addObject: scsiTarget];
                    [scsiTarget release];

                    switch ([[[scsiTarget unitWithUnitNumber: 0] unitDescription] peripheralType]) {
                        case OSCSIHardDrivePeripheralType:
                        case OSCSICDROMPeripheralType:
                        case OSCSIOpticalMemoryPeripheralType:
                        case OSCSIRAIDPeripheralType:
                            rawDeviceName = [NSString stringWithFormat: @"/dev/rsd%dh", diskTargets++];
                            break;
                        case OSCSITapePeripheralType:
                            rawDeviceName = [NSString stringWithFormat: @"/dev/rst%d", tapeTargets++];
                            break;
                        default:
                            rawDeviceName = nil;
                    }

                    [scsiTarget setRawDeviceName: rawDeviceName];
                }
            } NS_HANDLER {
                // Some error occured, log it and keep scanning
                NSLog(@"Error while examinging target %d on bus %d: %@",
                      (int)targetNumber, (int)_busNumber, localException);
            } NS_ENDHANDLER;
        }

        //
        // Search for ATAPI devices.  Under Rhapsody, these are the rsd devices
        // AFTER all of the SCSI devices.  We'll only do the search when we
        // are on bus zero since otherwise they'd show up on all busses.
        //
        if (_busNumber == 0) {
            while (YES) {
                NS_DURING {
                    // There is a device here
                    rawDeviceName = [NSString stringWithFormat: @"/dev/rsd%dh", diskTargets++];
                    scsiTarget = [[OSCSITarget alloc] initWithBusNumber: _busNumber
                                                           targetNumber: OSCSI_INVALID_TARGET_NUMBER
                                                          rawDeviceName: rawDeviceName];
                } NS_HANDLER {
                    // Some error occured, log it and keep scanning
                    NSLog(@"Error while examinging target %d on bus %d: %@",
                          targetNumber, _busNumber, localException);
                    scsiTarget = nil;
                } NS_ENDHANDLER;
                if (scsiTarget) {
                    [(NSMutableArray *)_targets addObject: scsiTarget];
                    [scsiTarget release];
                } else
                    break;
            }
        }
    }
    
    return _targets;
}

- (OSCSITarget *) targetWithTargetNumber: (unsigned int) targetNumber;
{
    unsigned int targetCount;

    // Make sure we've cached our targets
    [self targets];
    
    targetCount = [_targets count];
    while (targetCount--) {
        OSCSITarget *scsiTarget;

        scsiTarget = [_targets objectAtIndex: targetCount];
        if ([scsiTarget targetNumber] == targetNumber)
            return scsiTarget;
    }

    [NSException raise: NSInvalidArgumentException
                format: @"Target %d doesn't exist on bus %d", targetNumber, _busNumber];
    return nil;
}

- (OSCSITarget *) targetWithRawDeviceName: (NSString *) deviceName;
{
    unsigned int targetCount;

    // Make sure we've cached our targets
    [self targets];

    targetCount = [_targets count];
    while (targetCount--) {
        OSCSITarget *scsiTarget;
        NSString    *name;
        
        scsiTarget = [_targets objectAtIndex: targetCount];
        name = [scsiTarget rawDeviceName];
        if ([name isEqualToString: deviceName])
            return scsiTarget;
    }

    [NSException raise: NSInvalidArgumentException
                format: @"Device %@ doesn't exist on bus %d", deviceName, _busNumber];
    return nil;
}

- (void) reset;
{
    OBRequestConcreteImplementation(self, _cmd);
#if 0
    // We'd need to figure out what target/lun the SCSI controller lives on.  It might be better
    // to implement a subclass of OSCSIUnit for the controller device.
    if (ioctl(_busFD, SGIOCRST) < 0)
        [NSException raise: NSInternalInconsistencyException
                    format: @"Unable to reset SCSI bus, errno = %d", OMNI_ERRNO()];
#endif
}

- (NSArray *) allUnitsWithPeripheralType: (OSCSIPeripheralType) peripheralType;
{
    NSMutableArray *results;
    NSArray        *units;
    unsigned int    targetIndex, unitIndex;
    OSCSITarget    *target;
    OSCSIUnit      *unit;
    
    results = [NSMutableArray array];

    // Cache our targets
    [self targets];

    targetIndex = [_targets count];
    while (targetIndex--) {
        target = [_targets objectAtIndex: targetIndex];
        units = [target units];
        unitIndex = [units count];
        while (unitIndex--) {
            unit = [units objectAtIndex: unitIndex];
            if ([[unit unitDescription] peripheralType] == peripheralType)
                [results addObject: unit];
        }
    }

    return results;
}

- (NSMutableDictionary *)debugDictionary;
{
    NSMutableDictionary *dict;

    dict = [super debugDictionary];
    [dict setObject: [NSNumber numberWithInt: _busNumber] forKey: @"bus"];
    [dict setObject: [self targets] forKey: @"targets"];

    return dict;
}


@end

@implementation OSCSIBus (PrivateAPI)

- _initWithBusNumber: (OSCSIBusNumberType) busNumber;
{
    if (!(self = [super init]))
        return nil;

    _busNumber = busNumber;
    return self;
}

@end

