// Copyright 1998-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniSCSI/OSCSICDTableOfContents.h>

#import <OmniSCSI/OSCSICDProtocols.h>

@implementation OSCSICDTableOfContents

- initWithTableOfContentsDescriptions: (OSCSICDTableOfContentsEntry *) descriptions
                                count: (unsigned int) descriptionCount;
{
    if (!(self = [super init]))
        return nil;

    _descriptionCount  = descriptionCount;
    _trackDescriptions = descriptions;

    return self;
}

- (void) dealloc;
{
    NSZoneFree(NULL, _trackDescriptions);
    [super dealloc];
}

- (unsigned int) numberOfTracks;
{
    unsigned int maxTrack = 0;
    unsigned int descriptionIndex;

    for (descriptionIndex = 0; descriptionIndex < _descriptionCount; descriptionIndex++) {
        unsigned int trackNumber;

        trackNumber = _trackDescriptions[descriptionIndex].trackNumber;

        if (trackNumber >= 0xAA) {
            // This should be the lead-out area description
            continue;
        }

        if (trackNumber > maxTrack)
            maxTrack = trackNumber;
    }

    return maxTrack;
}

- (unsigned int) trackContainingLBA: (OSCSILogicalBlockAddress) lba;
{
    unsigned int descriptionIndex;

    for (descriptionIndex = 0; descriptionIndex < _descriptionCount; descriptionIndex++) {
        if (lba < _trackDescriptions[descriptionIndex].startingLBA) {
            if (descriptionIndex) {
                return _trackDescriptions[descriptionIndex-1].trackNumber;
            }
        }
    }

    [NSException raise: NSInvalidArgumentException
                format: @"No track contains LBA %d", lba];
    return 0;
}

- (OSCSILogicalBlockAddress) startingLBAForTrackNumber: (unsigned int) trackNumber;
{
    unsigned int descriptionIndex;

    for (descriptionIndex = 0; descriptionIndex < _descriptionCount; descriptionIndex++) {
        unsigned int number;

        number = _trackDescriptions[descriptionIndex].trackNumber;

        if (trackNumber == number)
            return _trackDescriptions[descriptionIndex].startingLBA;
    }

    [NSException raise: NSInvalidArgumentException
                format: @"No such track %d", trackNumber];
    return 0;
}

- (OSCSILogicalBlockAddress) endingLBAForTrackNumber: (unsigned int) trackNumber;
{
    unsigned int descriptionIndex;

    for (descriptionIndex = 0; descriptionIndex < _descriptionCount; descriptionIndex++) {
        unsigned int number;

        number = _trackDescriptions[descriptionIndex].trackNumber;

        if (trackNumber < number)
            return _trackDescriptions[descriptionIndex].startingLBA;
    }

    [NSException raise: NSInvalidArgumentException
                format: @"No such track %d", trackNumber];
    return 0;
}

- (NSMutableDictionary *) debugDictionary;
{
    NSMutableDictionary      *dict;
    NSMutableArray           *tracks;
    unsigned int              descriptionIndex;
    OSCSIMinuteSecondsFrames  absoluteMSF, relativeMSF;
    OSCSILogicalBlockAddress  lba;
    NSString                 *track;


    dict = [super debugDictionary];
    
    tracks = [[NSMutableArray alloc] init];
    [dict setObject: tracks forKey: @"tracks"];
    [tracks release];

    lba = _trackDescriptions[0].startingLBA;
    absoluteMSF = OSCSIAbsoluteLogicalBlockAddressToMinutesSecondsFrames(lba);
    relativeMSF = OSCSIRelativeLogicalBlockAddressToMinutesSecondsFrames(lba - 0);
    track = [NSString stringWithFormat: @"Track #%d, Absolute LBA = %d, Absolute MSF = %d:%d:%d, Relative MSF = %d:%d:%d",
        _trackDescriptions[0].trackNumber, lba,
        (int)absoluteMSF.minutes, (int)absoluteMSF.seconds, (int)absoluteMSF.frames,
        (int)relativeMSF.minutes, (int)relativeMSF.seconds, (int)relativeMSF.frames];
    [tracks addObject: track];
    
    for (descriptionIndex = 1; descriptionIndex < _descriptionCount; descriptionIndex++) {
        lba = _trackDescriptions[descriptionIndex].startingLBA;
        absoluteMSF = OSCSIAbsoluteLogicalBlockAddressToMinutesSecondsFrames(lba);
        relativeMSF = OSCSIRelativeLogicalBlockAddressToMinutesSecondsFrames(lba - _trackDescriptions[descriptionIndex-1].startingLBA);
            
        track = [NSString stringWithFormat: @"Track #%d, Absolute LBA = %d, Absolute MSF = %d:%d:%d, Relative MSF = %d:%d:%d",
            _trackDescriptions[descriptionIndex].trackNumber, lba,
            (int)absoluteMSF.minutes, (int)absoluteMSF.seconds, (int)absoluteMSF.frames,
            (int)relativeMSF.minutes, (int)relativeMSF.seconds, (int)relativeMSF.frames];
        [tracks addObject: track];
    }

    return dict;
}

@end
