// Copyright 1998-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniSCSI/OSCSIGeneric.h>

#import <OmniSCSI/OSCSIResponse.h>

#import <OmniBase/system.h>
#import <OmniBase/macros.h>
#import <dev/scsireg.h>
#import <dev/disk.h> // for DKIOCEJECT


static BOOL         *_usedDevices    = NULL;
static unsigned int  _deviceCount    = 0;
static unsigned int  _defaultTimeout = 60;   // one minute

NSString *OSCSINoAvailableDeviceException      = @"OSCSINoAvailableDeviceException";
NSString *OSCSIIOException                     = @"OSCSIIOException";
NSString *OSCSIBadRawDeviceException           = @"OSCSIBadRawDeviceException";
NSString *OSCSIGenericNoRawDeviceNameException = @"OSCSIGenericNoRawDeviceNameException";

NSString *OSCSINoSuchBusException     = @"OSCSINoSuchBusException";
NSString *OSCSINoSuchAddressException = @"OSCSINoSuchAddressException";

@interface OSCSIGeneric (PrivateAPI)
+ (int) _openGenericDevice: (int *) deviceIndex;
+ (void) _closeGenericDevice: (int) fd
                       index: (int) deviceIndex;
                               
- (OSCSIResponse *) _newResponseByPerformingSCSICommand: (struct scsi_req *) srp
                                                   name: (const char *) commandName
                                               withData: (NSData *) data;
@end

static BOOL OSCSIGenericDebug = NO;

@implementation OSCSIGeneric

+ (void) initialize;
{
    int          fd;
    char         path[MAXPATHLEN];

    [super initialize];
    if (_usedDevices)
        return;
    
    // Figure out how many sg devices are present
    _deviceCount = 0;
    do {
        sprintf(path, "/dev/sg%d", _deviceCount);
        fd = open(path, O_RDWR);
        if (fd > 0)
            close(fd);
        _deviceCount++;

        // Stop when we didn't get a file and it wasn't because someone else
        // is using it (which is a transient condition, really).
    } while (fd >= 0 || OMNI_ERRNO() == EBUSY);

    _deviceCount--;
    _usedDevices = (BOOL *)NSZoneCalloc(NULL, sizeof(BOOL), _deviceCount);


    OSCSIGenericDebug = [[OFUserDefaults sharedUserDefaults] boolForKey: @"OSCSIGenericDebug"];
    if (OSCSIGenericDebug)
        NSLog(@"OSCSIGenericDebug enabled");
}

+ (OSCSIBusNumberType) numberOfBusses;
{
    OSCSIBusNumberType busCount = 0;
    int                fd, index;

    fd = [self _openGenericDevice: &index];

    while (YES) {
        if (ioctl(fd, SGIOCCNTR, &busCount) < 0)
            break;
        busCount++;
    }

    [self _closeGenericDevice: fd index: index];
    return busCount;
}

+ (void) setDefaultTimeout: (unsigned int) timeout;
{
    _defaultTimeout = timeout;
}

+ (unsigned int) defaultTimeout;
{
    return _defaultTimeout;
}

+ (unsigned int) maximumTargetsPerBus;
{
    // Currently the SCSI-3 ioctls don't work, so we'll only search the smaller SCSI-2 space.
#warning Not searching all of the SCSI-3 space
    return SCSI_NTARGETS;
}

- initWithBusNumber: (OSCSIBusNumberType) busNumber
       targetNumber: (OSCSITargetNumberType) targetNumber
         unitNumber: (OSCSIUnitNumberType) unitNumber;
{
    _deviceIndex  = -1;
    _busNumber    = busNumber;
    _targetNumber = targetNumber;
    _timeout      = _defaultTimeout;
    _unitNumber   = unitNumber;
    _rawFD        = -1;
    _genericFD    = -1;
    _currentFD    = -1;
    
    if (_targetNumber != OSCSI_INVALID_TARGET_NUMBER) {
        scsi_adr_t address;
        
        // We use the 'unsigned long long' typedef for the target/unit, but the SCSI-3
        // ioctls apparently aren't working in OpenStep/4.2 or Rhapsody DR2.  We'll fall
        // back and use the small SCSI-2 versions.
        if (_targetNumber >= SCSI_NTARGETS) {
            [NSException raise: NSInvalidArgumentException
                        format: @"Target %d is bigger than the maximum of %d (due to a kernel bug).", _targetNumber, SCSI_NTARGETS];
        }
        if (_unitNumber >= SCSI_NLUNS) {
            [NSException raise: NSInvalidArgumentException
                        format: @"Unit %d is bigger than the maximum of %d (due to a kernel bug).", _unitNumber, SCSI_NLUNS];
        }
        

        NS_DURING {
            _genericFD = [isa _openGenericDevice: &_deviceIndex];
        } NS_HANDLER {
            [self release];
            [localException raise];
        } NS_ENDHANDLER;

        if (ioctl(_genericFD, SGIOCCNTR, &_busNumber) < 0) {
            close(_genericFD);
            [self release];
            [NSException raise: OSCSINoSuchBusException
                        format: @"Unable to select bus %d, errno = %d",  _busNumber, OMNI_ERRNO()];
        }

        // Attempt to set the address on the device
        address.sa_target = _targetNumber;
        address.sa_lun    = _unitNumber;

        if (ioctl(_genericFD, SGIOCSTL, &address) < 0) {
            close(_genericFD);
            [self release];
            [NSException raise: OSCSINoSuchAddressException
                        format: @"Unable to select the target/lun of %d/%d, errno = %d",
                _targetNumber, _unitNumber, OMNI_ERRNO()];
        }

        if (ioctl(_genericFD, SGIOCENAS) < 0) {
            close(_genericFD);
            [self release];
            [NSException raise: NSInternalInconsistencyException
                        format: @"Unable to set autosense on the bus/target/lun = %d/%d/%d, errno = %d",
                _busNumber, _targetNumber, _unitNumber, OMNI_ERRNO()];
        }
    }

    if (_deviceIndex < 0) {
        if (_targetNumber != OSCSI_INVALID_TARGET_NUMBER) {
            // No free sg device, or error.  Don't call -useRaw immediately in the ATAPI case
            // since that would cause a prompt for a disc to be inserted.
            [self release];
            [NSException raise: OSCSINoAvailableDeviceException
                        format: @"No generic SCSI device available.  Please release one that is in use."];
        }
    } else
        _currentFD = _genericFD;

    return self;
}

- (void) dealloc;
{
    if (_rawFD > 0)
        close(_rawFD);
    [isa _closeGenericDevice: _genericFD index: _deviceIndex];
    [_rawDeviceName release];
    [super dealloc];
}

- (void) setRawDeviceName: (NSString *) rawDeviceName;
{
    [_rawDeviceName release];
    _rawDeviceName = [rawDeviceName copy];
}

- (NSString *) rawDeviceName;
{
    return _rawDeviceName;
}

- (OSCSIBusNumberType) busNumber;
{
    return _busNumber;
}

- (OSCSITargetNumberType) targetNumber;
{
    return _targetNumber;
}

- (OSCSIUnitNumberType) unitNumber;
{
    return _unitNumber;
}

- (void) setTimeout: (unsigned int) timeout;
{
    _timeout = timeout;
}

- (unsigned int) timeout;
{
    return _timeout;
}

- (BOOL) useRaw;
{
    if (_rawFD < 0) {
        if (!_rawDeviceName)
            [NSException raise: OSCSIGenericNoRawDeviceNameException
                        format: @"%@ was not assigned a raw device name before -useRaw was called.", self];
        
        _rawFD = open([_rawDeviceName cString], O_RDWR);
    }

    if (_rawFD > 0) {
        _currentFD = _rawFD;
        return YES;
    } else
        return NO;
}

- (void) closeRaw;
{
    if (_rawFD < 0)
        return;
    
    if (_currentFD == _rawFD)
        _currentFD = _genericFD;

    close(_rawFD);
    _rawFD = -1;
}

- (BOOL) useGeneric
{
    if (_genericFD > 0) {
        _currentFD = _genericFD;
        return YES;
    } else
        return NO;
}


- (void) eject;
{
    // Make sure _rawFD is valid
    [self useRaw];
    
    if (ioctl(_rawFD, DKIOCEJECT) < 0)
        NSLog(@"Error in eject, errno = %d", OMNI_ERRNO());
    else
        [self closeRaw];
}

- (void) requestMedia;
{
    OSCSIResponse *response;
    BOOL           possibleMediaError = NO;
    int            attempt = 0;

    // Try this a couple of times -- NEXTSTEP has a bug where it
    // won't ask for media the first time you open a device after
    // a reboot.  Telling it to eject, closing and reopening
    // will make it work.
    while (attempt++ < 2) {
        if (![self useRaw]) {
            NSLog(@"Error opening raw device, errno = %d", OMNI_ERRNO());
            break;
        }

        NS_DURING {
            response = [self newResponseByWritingData: nil
                                               opcode: 0x00  // TEST UNIT READY
                                                 name: "TEST UNIT READY"
                                                 arg1: 0x0   // RESERVED
                                                 arg2: 0x0   // RESERVED
                                                 arg3: 0x0   // RESERVED
                                                 arg4: 0x0]; // RESERVED

        } NS_HANDLER {
            possibleMediaError = YES;
            response = nil;
        } NS_ENDHANDLER;

        if ([response isMediaMissingError])
            possibleMediaError = YES;
        [response release];

        // Check out the response value
        if (possibleMediaError) {
            NSLog(@"Attempting to open raw device again...");
            // Try again.
            //[self eject];
            [self closeRaw];
        } else
            break;
    }
        

    [self useGeneric];
}

//
// 6-byte commands
//
- (OSCSIResponse *) newResponseByWritingData: (NSData *) data
                                      opcode: (unsigned char) opcode
                                        name: (unsigned char *) name
                                        arg1: (unsigned char) arg1
                                        arg2: (unsigned char) arg2
                                        arg3: (unsigned char) arg3
                                        arg4: (unsigned char) arg4;
{
    struct scsi_req  sr;
    unsigned char   *command;

    memset(&sr, 0, sizeof(sr));
    command = (unsigned char *)&sr;
    
    sr.sr_cdb.cdb_opcode = opcode;
    command[1] = arg1;
    command[2] = arg2;
    command[3] = arg3;
    command[4] = arg4;
    command[5] = CTRL_NOLINK;

    sr.sr_dma_dir = SR_DMA_WR; // WRITE
    sr.sr_cdb_length = 6;
    return [self _newResponseByPerformingSCSICommand: &sr name: name withData: data];
}

- (OSCSIResponse *) newResponseByReadingData: (NSMutableData *) data
                                      opcode: (unsigned char) opcode
                                        name: (unsigned char *) name
                                        arg1: (unsigned char) arg1
                                        arg2: (unsigned char) arg2
                                        arg3: (unsigned char) arg3
                                        arg4: (unsigned char) arg4;
{
    struct scsi_req  sr;
    unsigned char   *command;

    memset(&sr, 0, sizeof(sr));
    command = (unsigned char *)&sr;

    sr.sr_cdb.cdb_opcode = opcode;
    command[1] = arg1;
    command[2] = arg2;
    command[3] = arg3;
    command[4] = arg4;
    command[5] = CTRL_NOLINK;

    sr.sr_dma_dir = SR_DMA_RD; // READ
    sr.sr_cdb_length = 6;
    return [self _newResponseByPerformingSCSICommand: &sr name: name withData: data];
}


// 10-byte commands
- (OSCSIResponse *) newResponseByWritingData: (NSData *) data
                                      opcode: (unsigned char) opcode
                                        name: (unsigned char *) name
                                        arg1: (unsigned char) arg1
                                        arg2: (unsigned char) arg2
                                        arg3: (unsigned char) arg3
                                        arg4: (unsigned char) arg4
                                        arg5: (unsigned char) arg5
                                        arg6: (unsigned char) arg6
                                        arg7: (unsigned char) arg7
                                        arg8: (unsigned char) arg8;
{
    struct scsi_req  sr;
    unsigned char   *command;

    memset(&sr, 0, sizeof(sr));
    command = (unsigned char *)&sr;

    sr.sr_cdb.cdb_opcode = opcode;
    command[1]  = arg1;
    command[2]  = arg2;
    command[3]  = arg3;
    command[4]  = arg4;
    command[5]  = arg5;
    command[6]  = arg6;
    command[7]  = arg7;
    command[8]  = arg8;
    command[9] = CTRL_NOLINK;

    sr.sr_dma_dir = SR_DMA_WR; // WRITE
    sr.sr_cdb_length = 10;
    return [self _newResponseByPerformingSCSICommand: &sr name: name withData: data];
}

- (OSCSIResponse *) newResponseByReadingData: (NSMutableData *) data
                                      opcode: (unsigned char) opcode
                                        name: (unsigned char *) name
                                        arg1: (unsigned char) arg1
                                        arg2: (unsigned char) arg2
                                        arg3: (unsigned char) arg3
                                        arg4: (unsigned char) arg4
                                        arg5: (unsigned char) arg5
                                        arg6: (unsigned char) arg6
                                        arg7: (unsigned char) arg7
                                        arg8: (unsigned char) arg8;
{
    struct scsi_req  sr;
    unsigned char   *command;

    memset(&sr, 0, sizeof(sr));
    command = (unsigned char *)&sr;

    sr.sr_cdb.cdb_opcode = opcode;
    command[1]  = arg1;
    command[2]  = arg2;
    command[3]  = arg3;
    command[4]  = arg4;
    command[5]  = arg5;
    command[6]  = arg6;
    command[7]  = arg7;
    command[8]  = arg8;
    command[9] = CTRL_NOLINK;

    sr.sr_dma_dir = SR_DMA_RD; // READ
    sr.sr_cdb_length = 10;
    return [self _newResponseByPerformingSCSICommand: &sr name: name withData: data];
}

    // 12-byte commands
- (OSCSIResponse *) newResponseByWritingData: (NSData *) data
                                      opcode: (unsigned char) opcode
                                        name: (unsigned char *) name
                                        arg1: (unsigned char) arg1
                                        arg2: (unsigned char) arg2
                                        arg3: (unsigned char) arg3
                                        arg4: (unsigned char) arg4
                                        arg5: (unsigned char) arg5
                                        arg6: (unsigned char) arg6
                                        arg7: (unsigned char) arg7
                                        arg8: (unsigned char) arg8
                                        arg9: (unsigned char) arg9
                                       arg10: (unsigned char) arg10;
{
    struct scsi_req  sr;
    unsigned char   *command;

    memset(&sr, 0, sizeof(sr));
    command = (unsigned char *)&sr;

    sr.sr_cdb.cdb_opcode = opcode;
    command[1]  = arg1;
    command[2]  = arg2;
    command[3]  = arg3;
    command[4]  = arg4;
    command[5]  = arg5;
    command[6]  = arg6;
    command[7]  = arg7;
    command[8]  = arg8;
    command[9]  = arg9;
    command[10] = arg10;
    command[11] = CTRL_NOLINK;

    sr.sr_dma_dir = SR_DMA_WR; // WRITE
    sr.sr_cdb_length = 12;
    return [self _newResponseByPerformingSCSICommand: &sr name: name withData: data];
}

- (OSCSIResponse *) newResponseByReadingData: (NSMutableData *) data
                                      opcode: (unsigned char) opcode
                                        name: (unsigned char *) name
                                        arg1: (unsigned char) arg1
                                        arg2: (unsigned char) arg2
                                        arg3: (unsigned char) arg3
                                        arg4: (unsigned char) arg4
                                        arg5: (unsigned char) arg5
                                        arg6: (unsigned char) arg6
                                        arg7: (unsigned char) arg7
                                        arg8: (unsigned char) arg8
                                        arg9: (unsigned char) arg9
                                       arg10: (unsigned char) arg10;
{
    struct scsi_req  sr;
    unsigned char   *command;

    memset(&sr, 0, sizeof(sr));
    command = (unsigned char *)&sr;

    sr.sr_cdb.cdb_opcode = opcode;
    command[1]  = arg1;
    command[2]  = arg2;
    command[3]  = arg3;
    command[4]  = arg4;
    command[5]  = arg5;
    command[6]  = arg6;
    command[7]  = arg7;
    command[8]  = arg8;
    command[9]  = arg9;
    command[10] = arg10;
    command[11] = CTRL_NOLINK;

    sr.sr_dma_dir = SR_DMA_RD; // READ
    sr.sr_cdb_length = 12;
    return [self _newResponseByPerformingSCSICommand: &sr name: name withData: data];
}

@end


@implementation OSCSIGeneric (PrivateAPI)

+ (int) _openGenericDevice: (int *) deviceIndex;
{
    unsigned int index;
    int          fd;
    
    // Search for an unused sg device
    for (index = 0; index < _deviceCount; index++) {
        if (!_usedDevices[index]) {
            char        path[MAXPATHLEN];

            sprintf(path, "/dev/sg%d", index);
            fd = open(path, O_RDWR);
            if (fd < 0) {
                if (OMNI_ERRNO() == EBUSY) {
                    // Someone is still using this device.  Try the next one.
                    continue;
                } else {
                    // We were able to open it in +initialize... something bad is going on.
                    [NSException raise: NSInternalInconsistencyException
                                format: @"Cannot open %s: %s", path, strerror(OMNI_ERRNO())];
                }
            }

            // success!
            _usedDevices[index] = YES;
            *deviceIndex = index;
            return fd;
        }
    }

    [NSException raise: OSCSINoAvailableDeviceException
                format: @"No generic SCSI device available.  Please release one that is in use."];
    return -1;
}

+ (void) _closeGenericDevice: (int) fd
                       index: (int) deviceIndex;
{
    if (fd > 0)
        close(fd);
    if (deviceIndex > 0)
        _usedDevices[deviceIndex] = NO;
}

- (OSCSIResponse *) _newResponseByPerformingSCSICommand: (struct scsi_req *) srp
                                                   name: (const char *) commandName
                                               withData: (NSData *) data;
{
    int                  rc;
    int                  bytesToTransfer; // don't trust that the field in the srp won't change
    unsigned int         byteIndex;
    const unsigned char *byte;
    

    OBPRECONDITION([data isKindOfClass: [NSMutableData class]] || srp->sr_dma_dir == SR_DMA_WR);
    OBPRECONDITION(srp->sr_cdb_length);
                   
    srp->sr_addr = (void *)[data bytes];
    srp->sr_dma_max = bytesToTransfer = [data length];
    srp->sr_ioto = _timeout;
    
    if (OSCSIGenericDebug) {
        fprintf(stderr, "\nCommand: %s\n", commandName);

        fprintf(stderr, "CDB (%d):", srp->sr_cdb_length);
        byte = (unsigned char *)&srp->sr_cdb;
        for (byteIndex = 0; byteIndex < srp->sr_cdb_length; byteIndex++)
            fprintf(stderr, " 0x%02x", (int)byte[byteIndex]);
        fprintf(stderr, "\n");

        if ((srp->sr_dma_dir == SR_DMA_WR) && bytesToTransfer) {
            unsigned int bytesToPrint;

            fprintf(stderr, "data dir = %d (%s)\n", srp->sr_dma_dir, (srp->sr_dma_dir == SR_DMA_RD) ? "SR_DMA_RD" : "SR_DMA_WR");
            fprintf(stderr, "data length = %d\n", srp->sr_dma_max);
            fprintf(stderr, "data        =");
            bytesToPrint = MIN(srp->sr_dma_max, 128);
            if (bytesToPrint) {
                byte = [data bytes];
                for (byteIndex = 0; byteIndex < bytesToPrint; byteIndex++)
                    fprintf(stderr, " 0x%02x", (int)byte[byteIndex]);
            }
            if (bytesToPrint < srp->sr_dma_max)
                fprintf(stderr, " ...\n");
            else
                fprintf(stderr, "\n");
        }
        
        fprintf(stderr, "timeout = %d\n", srp->sr_ioto);
    }
        
#warning SCSI-3 commands are not working in the current kernel

    if (_currentFD < 0) {
        // Instances allocated with only a _rawIndex won't automatically open
        // the raw device to avoid getting prompted to insert a disc.
        if (_genericFD > 0)
            [self useGeneric];
        else
            [self useRaw];
    }

#warning Make a OMNI_SET_ERRNO macro
    cthread_set_errno_self(0);
    rc = ioctl(_currentFD, SGIOCREQ, srp);
    if (OSCSIGenericDebug) {
        fprintf(stderr, "rc    = %d\n", rc);
        fprintf(stderr, "errno = %d\n", OMNI_ERRNO());
    }
    
    if (rc < 0) {
#warning Should OSCSIResponse encapsulate this?
        [NSException raise: OSCSIIOException
                    format: @"Error invoking SCSI command via ioctl(), errno = %d", OMNI_ERRNO()];
    }

    if (OSCSIGenericDebug) {
        fprintf(stderr, "scsi status byte = %d\n", srp->sr_scsi_status);
        fprintf(stderr, "driver status    = %d\n", srp->sr_io_status);
        fprintf(stderr, "bytes xfered     = %d (%d left)\n", srp->sr_dma_xfr, bytesToTransfer - srp->sr_dma_xfr);
        fprintf(stderr, "sense key                       = 0x%02x\n", srp->sr_esense.er_sensekey);
        fprintf(stderr, "additional sense code           = 0x%02x\n", srp->sr_esense.er_addsensecode);
        fprintf(stderr, "additional sense code qualifier = 0x%02x\n", srp->sr_esense.er_qualifier);

        if ((srp->sr_dma_dir == SR_DMA_RD) && bytesToTransfer) {
            unsigned int bytesToPrint;
            
            fprintf(stderr, "returned data =");
            bytesToPrint = MIN(srp->sr_dma_xfr, 128);
            if (bytesToPrint) {
                byte = [data bytes];
                for (byteIndex = 0; byteIndex < bytesToPrint; byteIndex++)
                    fprintf(stderr, " 0x%02x", (int)byte[byteIndex]);
            }
            if (bytesToPrint < srp->sr_dma_xfr)
                fprintf(stderr, " ...\n");
            else
                fprintf(stderr, "\n");
        }
    }

    if (srp->sr_scsi_status != STAT_GOOD ||
        srp->sr_io_status   != SR_IOST_GOOD ||
        srp->sr_dma_xfr     != bytesToTransfer) {
        return [[OSCSIResponse alloc] initWithStatusByte: srp->sr_scsi_status
                                            driverStatus: srp->sr_io_status
                                                senseKey: srp->sr_esense.er_sensekey
                                     additionalSenseCode: srp->sr_esense.er_addsensecode
                            additionalSenseCodeQualifier: srp->sr_esense.er_qualifier
                                      bytesNotTransfered: bytesToTransfer - srp->sr_dma_xfr];
    } else
        // Since we are defined to return a retained instance
        return [[OSCSIResponse goodResponse] retain];
}

@end
