// Copyright 1998-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniSCSI/OSCSITarget.h>

#import <OmniSCSI/OSCSIUnit.h>
#import <OmniSCSI/OSCSIGeneric.h>
#import <OmniSCSI/OSCSIUnitDescription.h>

@implementation OSCSITarget

- initWithBusNumber: (OSCSIBusNumberType) busNumber
       targetNumber: (OSCSITargetNumberType) targetNumber;
{
    OSCSIUnit            *scsiUnit;
    OSCSIUnitDescription *unitDescription = nil;
    OSCSIGeneric         *generic = nil;
    NSAutoreleasePool    *pool;
    
    if (!(self = [super init]))
        return nil;

    _busNumber    = busNumber;
    _targetNumber = targetNumber;

    // SCSI-3 SPC defines a 'REPORT LUNS' command, but there doesn't seem to be
    // anthing similar in SCSI-2.  Right now, we probably can't depend upon
    // all devices to implement 'REPORT LUNS'.

    // For the time being, we'll just assume that there is only one LUN.
#warning Determine available LUNs correctly

    NS_DURING {
        generic = [[OSCSIGeneric alloc] initWithBusNumber: _busNumber
                                             targetNumber: _targetNumber
                                               unitNumber: 0];
        unitDescription = [[OSCSIUnitDescription alloc] initWithGeneric: generic];
    } NS_HANDLER {
        if (![[localException name] isEqualToString: OSCSINoSuchAddressException])
            [localException raise];
        
        unitDescription = nil;
    } NS_ENDHANDLER;

    if (!unitDescription) {
        [generic release];
        [self release];
        return nil;
    }


    // Make sure that we don't leak the generic or even retain it longer
    // than we must.
    pool = [[NSAutoreleasePool alloc] init];
    NS_DURING {
        Class   unitClass;

        unitClass = [OSCSIUnit bestClassForUnitDescription: unitDescription];
        scsiUnit = [[unitClass alloc] initWithTarget: self
                                          unitNumber: 0
                                     unitDescription: unitDescription
                                    temporaryGeneric: generic];
    } NS_HANDLER {
        scsiUnit = nil;
        NSLog(@"Exception raised while initializing unit: %@", localException);
    } NS_ENDHANDLER;

    [generic release];
    [unitDescription release];
    [pool release];
    
    if (!scsiUnit) {
        [self release];
        return nil;
    }

    _units = [[NSMutableArray alloc] init];
    [(NSMutableArray *)_units addObject: scsiUnit];
    [scsiUnit release];
    
    return self;
}

- initWithBusNumber: (OSCSIBusNumberType) busNumber
       targetNumber: (OSCSITargetNumberType) targetNumber
      rawDeviceName: (NSString *) rawDeviceName;
{
    OSCSIUnit            *scsiUnit;
    OSCSIUnitDescription *unitDescription = nil;
    OSCSIGeneric         *generic = nil;
    NSAutoreleasePool    *pool;

    if (!(self = [super init]))
        return nil;

    _busNumber     = busNumber;
    _targetNumber  = targetNumber;
    _rawDeviceName = [rawDeviceName copy];
    
    // SCSI-3 SPC defines a 'REPORT LUNS' command, but there doesn't seem to be
    // anthing similar in SCSI-2.  Right now, we probably can't depend upon
    // all devices to implement 'REPORT LUNS'.

    // For the time being, we'll just assume that there is only one LUN.
#warning Determine available LUNs correctly

    NS_DURING {
        generic = [[OSCSIGeneric alloc] initWithBusNumber: busNumber
                                             targetNumber: _targetNumber
                                               unitNumber: 0];
        [generic setRawDeviceName: _rawDeviceName];
        unitDescription = [[OSCSIUnitDescription alloc] initWithGeneric: generic];
    } NS_HANDLER {
        unitDescription = nil;
    } NS_ENDHANDLER;

    if (!unitDescription) {
        [generic release];
        [self release];
        return nil;
    }

    // Make sure that we don't leak the generic or even retain it longer
    // than we must.
    pool = [[NSAutoreleasePool alloc] init];
    NS_DURING {
        Class   unitClass;

        unitClass = [OSCSIUnit bestClassForUnitDescription: unitDescription];
        scsiUnit = [[unitClass alloc] initWithTarget: self unitNumber: 0 unitDescription: unitDescription temporaryGeneric: generic];
    } NS_HANDLER {
        scsiUnit = nil;
        NSLog(@"Exception raised while initializing unit: %@", localException);
    } NS_ENDHANDLER;

    [generic release];
    [unitDescription release];
    [pool release];
    
    if (!scsiUnit) {
        [self release];
        return nil;
    }

    _units = [[NSMutableArray alloc] init];
    [(NSMutableArray *)_units addObject: scsiUnit];
    [scsiUnit release];

    return self;
}

- (void) dealloc;
{
    [_units release];
    [_rawDeviceName release];
    [super dealloc];
}

- (void) setRawDeviceName: (NSString *) rawDeviceName;
{
    [_rawDeviceName release];
    _rawDeviceName = [rawDeviceName copy];
}

- (NSString *) rawDeviceName;
{
    return _rawDeviceName;
}

- (OSCSIBusNumberType) busNumber;
{
    return _busNumber;
}

- (OSCSITargetNumberType) targetNumber;
{
    return _targetNumber;
}

- (NSArray *) units;
{
    return _units;
}

- (OSCSIUnit *) unitWithUnitNumber: (OSCSIUnitNumberType) unitNumber;
{
    unsigned int unitCount;

    unitCount = [_units count];
    while (unitCount--) {
        OSCSIUnit *scsiUnit;

        scsiUnit = [_units objectAtIndex: unitCount];
        if ([scsiUnit unitNumber] == unitNumber)
            return scsiUnit;
    }

    [NSException raise: NSInvalidArgumentException
                format: @"Unit %d doesn't exist on target %d", unitNumber, _targetNumber];
    return nil;
}

- (NSMutableDictionary *)debugDictionary;
{
    NSMutableDictionary *dict;
    NSMutableArray      *unitDicts;
    NSArray             *units;
    unsigned int         unitIndex, unitCount;
    OSCSIUnit           *unit;
    
    dict = [super debugDictionary];
    [dict setObject: [NSNumber numberWithInt: _targetNumber] forKey: @"target"];
    [dict setObject: _rawDeviceName ? _rawDeviceName : @"Unknown" forKey: @"rawDeviceName"];

    unitDicts = [NSMutableArray array];
    units = [self units];
    unitCount = [units count];

    for (unitIndex = 0; unitIndex < unitCount; unitIndex++) {
        unit = [units objectAtIndex: unitIndex];
        [unitDicts addObject: [unit debugDictionary]];
    }

    [dict setObject: unitDicts forKey: @"units"];

    return dict;
}

@end
