// Copyright 1998-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniSCSI/OSCSIUnit.h>

#import <OmniSCSI/OSCSIGeneric.h>
#import <OmniSCSI/OSCSIUnitDescription.h>
#import <OmniSCSI/OSCSIResponse.h>

#import <dev/scsireg.h>

#import <OmniFoundation/NSString-OFExtensions.h>

static NSMutableArray *mappings = nil;

static NSString *_OSCSIUnitPeripheralTypeKey = @"peripheralType";
static NSString *_OSCSIUnitVendorNameKey     = @"vendorName";
static NSString *_OSCSIUnitProductNameKey    = @"productName";
static NSString *_OSCSIUnitRevisionKey       = @"revision";
static NSString *_OSCSIResponseFormatKey     = @"responseFormat";

static BOOL OSCSIUnitMatchingDebug = NO;

@implementation OSCSIUnit

+ (void) initialize;
{
    [super initialize];
    if (mappings)
        return;

    OSCSIUnitMatchingDebug = [[OFUserDefaults sharedUserDefaults] boolForKey: @"OSCSIUnitMatchingDebug"];
    if (OSCSIUnitMatchingDebug)
        NSLog(@"OSCSIUnitMatchingDebug enabled");
    
    mappings = [[NSMutableArray alloc] init];
}

// OFBundleRegistryTarget informal protocol
+ (void)registerItemName:(NSString *)itemName bundle:(NSBundle *)bundle description:(NSDictionary *)description;
{
    unsigned int         mappingCount;
    NSMutableDictionary *mapping;
    
    if (![itemName isEqualToString: @"deviceMappings"])
        return;

    if (OSCSIUnitMatchingDebug) {
        NSLog(@"Registering %@ %@", itemName, description);
    }

    mappingCount = [(NSArray *)description count];
    while (mappingCount--) {
        NSString            *pattern;
        OFRegularExpression *expression;
        
        mapping = [[[(NSArray *)description objectAtIndex: mappingCount] mutableCopy] autorelease];

        //
        // Cache the compiled regular expression patterns
        //

        // _OSCSIUnitPeripheralTypeKey
        if ((pattern = [mapping objectForKey: _OSCSIUnitPeripheralTypeKey])) {
            expression = [[OFRegularExpression alloc] initWithString: pattern];
            [mapping setObject: expression forKey: _OSCSIUnitPeripheralTypeKey];
            [expression release];
        }

        // _OSCSIUnitVendorNameKey
        if ((pattern = [mapping objectForKey: _OSCSIUnitVendorNameKey])) {
            expression = [[OFRegularExpression alloc] initWithString: pattern];
            [mapping setObject: expression forKey: _OSCSIUnitVendorNameKey];
            [expression release];
        }

        // _OSCSIUnitProductNameKey
        if ((pattern = [mapping objectForKey: _OSCSIUnitProductNameKey])) {
            expression = [[OFRegularExpression alloc] initWithString: pattern];
            [mapping setObject: expression forKey: _OSCSIUnitProductNameKey];
            [expression release];
        }

        // _OSCSIUnitRevisionKey
        if ((pattern = [mapping objectForKey: _OSCSIUnitRevisionKey])) {
            expression = [[OFRegularExpression alloc] initWithString: pattern];
            [mapping setObject: expression forKey: _OSCSIUnitRevisionKey];
            [expression release];
        }

        // _OSCSIResponseFormatKey
        if ((pattern = [mapping objectForKey: _OSCSIResponseFormatKey])) {
            expression = [[OFRegularExpression alloc] initWithString: pattern];
            [mapping setObject: expression forKey: _OSCSIResponseFormatKey];
            [expression release];
        }

        // Put the bundle in the mapping dictionary.  We'll use it later to find the class (if it isn't loaded already).
        [mapping setObject: bundle forKey: @"bundle"];

        // Save the mapping
        [mappings addObject: mapping];
    }
}

static int _OSCSIOrderMappingByPriority(id object1, id object2, void *context)
{
    int priority1, priority2;

    priority1 = [[object1 objectForKey: @"priority"] intValue];
    priority2 = [[object2 objectForKey: @"priority"] intValue];

    return priority1 - priority2;
}

+ (NSArray *) classesForUnitDescription: (OSCSIUnitDescription *) unitDescription;
{
    unsigned int    mappingIndex;
    NSMutableArray *candidates;
    NSMutableArray *classes;
    NSDictionary   *mapping;
    
    candidates = [[NSMutableArray alloc] init];

    if (OSCSIUnitMatchingDebug)
        NSLog(@"looking for matches for unitDescription = %@ from mappings %@", unitDescription, mappings);
    
    // Loop through all of the patterns, testing each one against the input description
    mappingIndex = [mappings count];
    if (!mappingIndex) {
        [NSException raise: @"OSCSINoMappings"
                    format: @"No mappings have been registered.  Did you call +[OBPostLoader processClasses]?"];
    }
    
    while (mappingIndex--) {
        OFRegularExpression      *expression;
        OFRegularExpressionMatch *match;
        
        mapping = [mappings objectAtIndex: mappingIndex];

        if (OSCSIUnitMatchingDebug)
            NSLog(@"mapping = %@", mapping);

        if ((expression = [mapping objectForKey: _OSCSIUnitPeripheralTypeKey])) {
            match = [expression matchInString: [unitDescription peripheralTypeName]];
            if (OSCSIUnitMatchingDebug)
                NSLog(@"peripheralTypeName match = %@", match, [unitDescription peripheralTypeName]);
            if (!match)
                continue;
        }

        if ((expression = [mapping objectForKey: _OSCSIUnitVendorNameKey])) {
            match = [expression matchInString: [unitDescription vendorIdentification]];
            if (OSCSIUnitMatchingDebug)
                NSLog(@"vendorIdentification match = %@", match);
            if (!match)
                continue;
        }

        if ((expression = [mapping objectForKey: _OSCSIUnitProductNameKey])) {
            match = [expression matchInString: [unitDescription productIdentification]];
            if (OSCSIUnitMatchingDebug)
                NSLog(@"productIdentification match = %@", match);
            if (!match)
                continue;
        }

        if ((expression = [mapping objectForKey: _OSCSIUnitRevisionKey])) {
            match = [expression matchInString: [unitDescription productRevisionLevel]];
            if (OSCSIUnitMatchingDebug)
                NSLog(@"productRevisionLevel match = %@", match);
            if (!match)
                continue;
        }

        if ((expression = [mapping objectForKey: _OSCSIResponseFormatKey])) {
            NSString *responseFormat;

            responseFormat = [NSString stringWithFormat: @"%d", [unitDescription responseFormat]];
            if (OSCSIUnitMatchingDebug)
                NSLog(@"responseFormat = %@", responseFormat);
            match = [expression matchInString: responseFormat];
            if (OSCSIUnitMatchingDebug)
                NSLog(@"responseFormat match = %@", match);
            if (!match)
                continue;
        }


        if (OSCSIUnitMatchingDebug)
            NSLog(@"added mapping as candidate");
        [candidates addObject: mapping];
    }

    // Sort the remaining candidates by priority
    [candidates sortUsingFunction: _OSCSIOrderMappingByPriority context: NULL];
    if (OSCSIUnitMatchingDebug)
        NSLog(@"sorted candidates = %@", candidates);

    
    classes = [NSMutableArray array];
    mappingIndex = [candidates count];
    while (mappingIndex--) {
        Class     mappingClass;
        NSString *className;
        
        mapping = [candidates objectAtIndex: mappingIndex];
        className = [mapping objectForKey: @"className"];
        mappingClass = NSClassFromString(className);
        if (!mappingClass) {
            NSBundle *bundle;
            
            // Try loading the bundle
            bundle = [mapping objectForKey: @"bundle"];
            mappingClass = [bundle classNamed: className];
            if (!mappingClass) {
                NSLog(@"Cannot find the OSCSIUnit class %@ that was supposed to reside in the bundle %@", className, bundle);
                continue;
            }
        }

        [classes addObject: mappingClass];
    }
    [candidates release];
    
    return classes;
}

+ (Class) bestClassForUnitDescription: (OSCSIUnitDescription *) unitDescription;
{
    NSArray *classes;

    classes = [self classesForUnitDescription: unitDescription];
    if ([classes count])
        return [classes objectAtIndex: 0];
    else {
        NSLog(@"No suitable mapping found for unit description %@", [unitDescription debugDictionary]);
        return nil;
    }
}

+ (NSString *) displayName;
{
    return @"Generic SCSI Unit";
}

- initWithTarget: (OSCSITarget *) target
      unitNumber: (OSCSIUnitNumberType) unitNumber
 unitDescription: (OSCSIUnitDescription *) unitDescription
temporaryGeneric: (OSCSIGeneric *) temporaryGeneric;
{
    if (!(self = [super init]))
        return nil;

    _nonretained_target = target;
    _unitNumber = unitNumber;
    _unitDescription = [unitDescription retain];

    if (![_unitDescription mediaIsRemovable])
        // Then is damn well better be there!
        _mediaPresent = YES;
    
    return self;
}

- (void) dealloc;
{
    [self close];
    [_unitDescription release];
    [super dealloc];
}

- (void) open;
{
    _generic = [[OSCSIGeneric alloc] initWithBusNumber: [_nonretained_target busNumber]
                                          targetNumber: [_nonretained_target targetNumber]
                                            unitNumber: _unitNumber];
    [_generic setRawDeviceName: [_nonretained_target rawDeviceName]];
}

- (void) close;
{
    [_generic release];
    _generic = nil;
}

- (BOOL) isOpen;
{
    return _generic != nil;
}

- (OSCSITarget *) target;
{
    return _nonretained_target;
}

- (OSCSIUnitNumberType) unitNumber;
{
    return _unitNumber;
}

- (OSCSIUnitDescription *) unitDescription;
{
    return _unitDescription;
}

- (OSCSIResponse *) testUnitReady;
{
    OSCSIResponse *response;

    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: nil
                                           opcode: 0x00  // TEST UNIT READY
                                             name: "TEST UNIT READY"
                                             arg1: 0x0   // RESERVED
                                             arg2: 0x0   // RESERVED
                                             arg3: 0x0   // RESERVED
                                             arg4: 0x0]; // RESERVED

    if ([response isError]) {
        BOOL           mediaMissingError = NO;

        if ([_unitDescription mediaIsRemovable]) {
            if ([response isMediaMissingError])
                mediaMissingError = YES;
        }

        if (mediaMissingError) {
            if (_mediaPresent)
                [self mediaRemoved];
        } else
            NSLog(@"Error in _testUnitReady - %@", [response errorString]);
    } else if (!_mediaPresent)
        [self mediaInserted];

    return [response autorelease];
}

- (OSCSIResponse *) requestSense;
{
    OSCSIResponse *response;
    NSMutableData *senseData;

    senseData = [[NSMutableData alloc] initWithLength: sizeof(esense_reply_t)];
    OBASSERT(_generic);
    response = [_generic newResponseByReadingData: senseData
                                           opcode: 0x03  // REQUEST SENSE
                                             name: "REQUEST SENSE"
                                             arg1: 0x0   // RESERVED
                                             arg2: 0x0   // RESERVED
                                             arg3: 0x0   // RESERVED
                                             arg4: sizeof(esense_reply_t)];
    if ([response isError])
        NSLog(@"Error requesting sense data - %@", [response errorString]);
    [senseData release];

    // Really we would like to make sure that the response has the same
    // data as what it is the senseData.
    return [response autorelease];
}

- (BOOL) mediaIsPresent;
{
    return _mediaPresent;
}

- (void) mediaInserted;
{
    _mediaPresent = YES;
}

- (void) mediaRemoved;
{
    _mediaPresent = NO;
}

- (void) setAllowsMediaRemoval: (BOOL) yn;
{
    OSCSIResponse *response;
    
    if (![_unitDescription mediaIsRemovable])
        return;

    OBASSERT(_generic);
    response = [_generic newResponseByWritingData: nil
                                           opcode: 0x1E
                                             name: "PREVENT/ALLOW MEDIUM REMOVAL"
                                             arg1: 0x0   // RESERVED
                                             arg2: 0x0   // RESERVED
                                             arg3: 0x0   // RESERVED
                                             arg4: yn ? 0x0 : 0x1]; // Prevent

    if ([response isError])
        NSLog(@"Error in %s - %@", _cmd, [response errorString]);

    [response release];
}

- (void) requestMedia;
{
    if (![_unitDescription mediaIsRemovable])
        return;

    [self testUnitReady];
    if ([self mediaIsPresent])
        return;

    OBASSERT(_generic);
    [_generic requestMedia];

    // Run the test again to send out the -mediaLoaded message
    [self testUnitReady];
}

- (void) eject;
{
    OBASSERT(_generic);
    [_generic eject];

    // Don't call -testUnitReady since it would cause the device to be opened again.
    // This is bad since if we are using the raw device only, we'll get prompted for
    // media.
    if ([_unitDescription mediaIsRemovable] && [self mediaIsPresent])
        [self mediaRemoved];
}

- (NSMutableDictionary *)debugDictionary;
{
    NSMutableDictionary *dict;

    dict = [super debugDictionary];
    [dict setObject: [_unitDescription debugDictionary] forKey: @"unitDescription"];
    [dict setObject: _generic ? @"Y" : @"N" forKey: @"isOpen"];
    if (_generic)
        [dict setObject: _generic forKey: @"_generic"];

    return dict;
}

@end
