// Copyright 1998-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <OmniSCSI/OSCSIUnitDescription.h>

#import <OmniFoundation/OmniFoundation.h>

#import <OmniSCSI/OSCSIGeneric.h>
#import <OmniSCSI/OSCSIResponse.h>

#import <dev/scsireg.h>  // Currently only used for inquire_reply_t

static NSString *defaultVendorIdentification = @"UnknownVendor";
static NSString *defaultProductIdentification = @"UnknownProduct";
static NSString *defaultProductRevision = @"UnknownRevision";

static NSString *_peripheralTypeNames[] = {
    @"Hard Drive",
    @"Tape",
    @"Printer",
    @"Processor",
    @"WORM",
    @"CDROM",
    @"Scanner",
    @"Optical Memory",
    @"Medium Changer",
    @"Network",
    @"Unknown (0x0A)",
    @"Unknown (0x0B)",
    @"RAID",
    @"Enclosure"
};

// Older versions
#define OSCSI_UNIT_DESCRIPTION_VERSION_19980710 (0)
#define OSCSI_UNIT_DESCRIPTION_VERSION_19980813 (1)

// The curent version
#define OSCSI_UNIT_DESCRIPTION_VERSION (OSCSI_UNIT_DESCRIPTION_VERSION_19980813)

@implementation OSCSIUnitDescription

+ (void)initialize;
{
    static BOOL initialized = NO;

    [super initialize];
    if (initialized)
        return;

    initialized = YES;
    [self setVersion:OSCSI_UNIT_DESCRIPTION_VERSION];
}

- initWithGeneric:(OSCSIGeneric *)generic;
{
    NSString *tmpString;
    NSMutableData *inquiryData;
    inquiry_reply_t *inquiryReply;
    OSCSIResponse *response;

    // This works for at least some ATAPI devices although the ATAPI spec says that this
    // should use a 12-byte command.
    inquiryData = [[NSMutableData alloc] initWithLength:sizeof(inquiry_reply_t)];
    response = [generic newResponseByReadingData:inquiryData
                                          opcode:0x12 // INQUIRY
                                            name:"INQUIRY"
                                            arg1:0x0
                                            arg2:0x0
                                            arg3:0x0
                                            arg4:[inquiryData length]];

    if ([response isError]) {
        // No device here
        [self release];
        return nil;
    }

    inquiryReply = [inquiryData mutableBytes];
    _peripheralType = (OSCSIPeripheralType)inquiryReply->ir_devicetype;
    _mediaIsRemovable = (inquiryReply->ir_removable != 0);
    _version = (OSCSIVersion)inquiryReply->ir_ansiversion;
    _responseFormat = (OSCSIVersion)inquiryReply->ir_rspdatafmt;

    tmpString = [[NSString alloc] initWithCString:inquiryReply->ir_vendorid length:sizeof(inquiryReply->ir_vendorid)];
    _vendorIdentification = [[tmpString stringByRemovingSurroundingWhitespace] retain];
    [tmpString release];

    tmpString = [[NSString alloc] initWithCString:inquiryReply->ir_productid length:sizeof(inquiryReply->ir_productid)];
    _productIdentification = [[tmpString stringByRemovingSurroundingWhitespace] retain];
    [tmpString release];

    tmpString = [[NSString alloc] initWithCString: inquiryReply->ir_revision length:sizeof(inquiryReply->ir_revision)];
    _productRevisionLevel = [[tmpString stringByRemovingSurroundingWhitespace] retain];
    [tmpString release];

    [inquiryData release];

    return self;
}

- (id)initWithArchivePropertyList:(id)propertyList;
{
    if (![super init])
        return nil;
    
    _vendorIdentification = [[propertyList objectForKey:@"vendorIdentification"] retain];
    _productIdentification = [[propertyList objectForKey:@"productIdentification"] retain];
    _productRevisionLevel = [[propertyList objectForKey:@"productRevision"] retain];
    _mediaIsRemovable = [propertyList boolForKey:@"mediaIsRemovable" defaultValue:YES];
    _peripheralType = [propertyList intForKey:@"peripheralType" defaultValue:0];
    _version = [propertyList intForKey:@"version" defaultValue:OSCSI3Version];
    _responseFormat = [propertyList intForKey:@"responseFormat" defaultValue:OSCSI3Version];

    return self;
}

- (void)dealloc;
{
    [_vendorIdentification release];
    [_productIdentification release];
    [_productRevisionLevel release];
    [super dealloc];
}

- (OSCSIPeripheralType) peripheralType;
{
    return _peripheralType;
}

- (BOOL) mediaIsRemovable;
{
    return _mediaIsRemovable;
}

- (OSCSIVersion)version;
{
    return _version;
}

- (OSCSIResponseFormat) responseFormat;
{
    return _responseFormat;
}

- (NSString *)peripheralTypeName;
{
    unsigned int index;

    index = (unsigned int)[self peripheralType];
    if (index < (sizeof(_peripheralTypeNames) / sizeof(*_peripheralTypeNames))) {
        return _peripheralTypeNames[index];
    } else
        return @"Unknown Peripheral Type";
}

- (NSString *)vendorIdentification;
{
    return _vendorIdentification ? _vendorIdentification : defaultVendorIdentification;
}

- (NSString *)productIdentification;
{
    return _productIdentification ? _productIdentification : defaultProductIdentification;
}

- (NSString *)productRevisionLevel;
{
    return _productRevisionLevel ? _productRevisionLevel : defaultProductRevision;
}

- (NSString *)unitDisplayName;
{
    return [NSString stringWithFormat: @"%@ %@ %@", _vendorIdentification, _productIdentification, _productRevisionLevel];
}

- (OSCSIUnitComparisonResult)compareToUnitDescription:(OSCSIUnitDescription *)unitDescription;
{
    if (![_vendorIdentification isEqualToString:[unitDescription vendorIdentification]])
        return OSCSIDifferentUnit;

    if (![_productIdentification isEqualToString:[unitDescription productIdentification]])
        return OSCSISameUnitVendor;

    if (![_productRevisionLevel isEqualToString:[unitDescription productRevisionLevel]])
        return OSCSISameUnitProduct;

    return OSCSISameUnit;
}


//
// Archiving
//

- (id)archivePropertyList;
{
    NSMutableDictionary *archiveDictionary;

    archiveDictionary = [[NSMutableDictionary alloc] init];

    [archiveDictionary setIntValue:_peripheralType forKey:@"peripheralType" defaultValue:0];
    [archiveDictionary setBoolValue:_mediaIsRemovable forKey:@"mediaIsRemovable" defaultValue:YES];
    [archiveDictionary setIntValue:_version forKey:@"version" defaultValue:OSCSI3Version];
    [archiveDictionary setIntValue:_responseFormat forKey:@"responseFormat" defaultValue:OSCSI3Version];
    [archiveDictionary setObject:[self vendorIdentification] forKey:@"vendorIdentification" defaultObject:defaultVendorIdentification];
    [archiveDictionary setObject:[self productIdentification] forKey:@"productIdentification" defaultObject:defaultProductIdentification];
    [archiveDictionary setObject:[self productRevisionLevel] forKey:@"productRevision" defaultObject:defaultProductRevision];

    return archiveDictionary;
}

//
// Subclassed from superclass
//

- (NSMutableDictionary *)debugDictionary;
{
    NSMutableDictionary *dict;

    dict = [super debugDictionary];
    [dict setObject: [self peripheralTypeName] forKey: @"peripheralType"];
    [dict setObject: _mediaIsRemovable ? @"Y" : @"N" forKey: @"mediaIsRemovable"];
    [dict setObject: [NSNumber numberWithInt: _version] forKey: @"version"];
    [dict setObject: [NSNumber numberWithInt: _responseFormat] forKey: @"responseFormat"];
    [dict setObject: [self vendorIdentification] forKey: @"vendorIdentification"];
    [dict setObject: [self productIdentification] forKey: @"productIdentification"];
    [dict setObject: [self productRevisionLevel] forKey: @"productRevision"];

    return dict;
}

//
// NSCoding implementation
//

- (void) encodeWithCoder: (NSCoder *) coder;
{
    [coder encodeObject: _vendorIdentification];
    [coder encodeObject: _productIdentification];
    [coder encodeObject: _productRevisionLevel];
    [coder encodeValueOfObjCType: @encode(typeof(_mediaIsRemovable)) at: &_mediaIsRemovable];
    [coder encodeValueOfObjCType: @encode(typeof(_peripheralType)) at: &_peripheralType];
    [coder encodeValueOfObjCType: @encode(typeof(_version)) at: &_version];
    [coder encodeValueOfObjCType: @encode(typeof(_responseFormat)) at: &_responseFormat];
}

- (id) initWithCoder: (NSCoder *) coder;
{
    unsigned int classVersion;

    // Don't use NSStringFromClass(isa) since it is slow (autoreleases a string).
    classVersion = [coder versionForClassName: @"OSCSIUnitDescription"];
    
    // There is only one version currently... no switch on version necessary.
    _vendorIdentification  = [[coder decodeObject] retain];
    _productIdentification = [[coder decodeObject] retain];
    _productRevisionLevel  = [[coder decodeObject] retain];
    [coder decodeValueOfObjCType: @encode(typeof(_mediaIsRemovable)) at: &_mediaIsRemovable];
    [coder decodeValueOfObjCType: @encode(typeof(_peripheralType)) at: &_peripheralType];
    [coder decodeValueOfObjCType: @encode(typeof(_version)) at: &_version];

    if (classVersion > OSCSI_UNIT_DESCRIPTION_VERSION_19980710) {
        [coder decodeValueOfObjCType: @encode(typeof(_responseFormat)) at: &_responseFormat];
    }

    return self;
}

//
// NSObject subclass
//

- (BOOL)isEqual:(id)object;
{
    OSCSIUnitDescription *other = object;

    if (!other)
        return NO;

    if (other == self)
        return YES;

    return [_vendorIdentification isEqual:other->_vendorIdentification] && [_productIdentification isEqual:other->_productIdentification] && [_productRevisionLevel isEqual:other->_productRevisionLevel] && (_mediaIsRemovable == other->_mediaIsRemovable) && (_peripheralType == other->_peripheralType) && (_version == other->_version) && (_responseFormat == other->_responseFormat);
}

@end
