// Copyright 1998-2000 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// http://www.omnigroup.com/DeveloperResources/OmniSourceLicense.html.

#import <Foundation/Foundation.h>
#import <OmniSCSI/OmniSCSI.h>
#import <OmniBase/OmniBase.h>
#import <OmniBase/system.h>

#ifdef RHAPSODY
#import <SoundKitHeaders/soundstruct.h>
#else
#import <SoundKit/soundstruct.h>
#endif

const char *pgm = NULL;

static void writeSNDHeader(int fd, unsigned int byteCount)
{
    SNDSoundStruct snd;
    int            rc;
    
    snd.magic        = SND_MAGIC;
    snd.dataLocation = sizeof(snd);
    snd.dataSize     = byteCount;
    snd.dataFormat   = SND_FORMAT_LINEAR_16;
    snd.samplingRate = SND_RATE_HIGH;
    snd.channelCount = 2;
    snd.info[0]      = '\0';
    snd.info[1]      = '\0';
    snd.info[2]      = '\0';
    snd.info[3]      = '\0';

    rc = write(fd, &snd, sizeof(snd));
    if (rc != sizeof(snd)) {
        fprintf(stderr, "%s: Unable to write snd header.\n", pgm);
        perror("write");
        exit(1);
    }
}

#define WAVE_FORMAT_PCM 1
/* MMIO macros */
#define mmioFOURCC(ch0, ch1, ch2, ch3) \
((unsigned int)(unsigned char)(ch0) | ((unsigned int)(unsigned char)(ch1) << 8) | \
 ((unsigned int)(unsigned char)(ch2) << 16) | ((unsigned int)(unsigned char)(ch3) << 24))

#define FOURCC_RIFF	mmioFOURCC ('R', 'I', 'F', 'F')
#define FOURCC_LIST	mmioFOURCC ('L', 'I', 'S', 'T')
#define FOURCC_WAVE	mmioFOURCC ('W', 'A', 'V', 'E')
#define FOURCC_FMT	mmioFOURCC ('f', 'm', 't', ' ')
#define FOURCC_DATA	mmioFOURCC ('d', 'a', 't', 'a')

static void writeWAVHeader(int fd, unsigned int byteCount)
{
    typedef unsigned int FOURCC;	/* a four character code */

    typedef struct CHUNKHDR {
        FOURCC ckid;		/* chunk ID */
        unsigned int dwSize; 	/* chunk size */
    } CHUNKHDR;

    /* flags for 'wFormatTag' field of WAVEFORMAT */

    /* specific waveform format structure for PCM data */
    typedef struct pcmwaveformat_tag {
        unsigned short wFormatTag;	/* format type */
        unsigned short nChannels;	/* number of channels (i.e. mono, stereo, etc.) */
        unsigned int nSamplesPerSec; /* sample rate */
        unsigned int nAvgBytesPerSec; /* for buffer size estimate */
        unsigned short nBlockAlign;	/* block size of data */
        unsigned short wBitsPerSample;
    } PCMWAVEFORMAT;
    typedef PCMWAVEFORMAT *PPCMWAVEFORMAT;

    /* simplified Header for standard WAV files */
    typedef struct WAVEHDR {
        CHUNKHDR chkRiff;
        FOURCC fccWave;
        CHUNKHDR chkFmt;
        unsigned short wFormatTag;	/* format type */
        unsigned short nChannels;	/* number of channels (i.e. mono, stereo, etc.) */
        unsigned int nSamplesPerSec; /* sample rate */
        unsigned int nAvgBytesPerSec; /* for buffer estimation */
        unsigned short nBlockAlign;	/* block size of data */
        unsigned short wBitsPerSample;
        CHUNKHDR chkData;
    } WAVEHDR;
    WAVEHDR waveHdr;

    unsigned long channels = 2;
    unsigned long nBitsPerSample = 16;
    unsigned long rate = 44100;
    
    unsigned long nBlockAlign = channels * ((nBitsPerSample + 7) / 8);
    unsigned long nAvgBytesPerSec = nBlockAlign * rate;
    int                 rc;

    waveHdr.chkRiff.ckid    = NSSwapHostLongToLittle(FOURCC_RIFF);
    waveHdr.chkRiff.dwSize  = NSSwapHostLongToLittle(byteCount + sizeof(WAVEHDR) - sizeof(CHUNKHDR));
    waveHdr.fccWave         = NSSwapHostLongToLittle(FOURCC_WAVE);
    waveHdr.chkFmt.ckid     = NSSwapHostLongToLittle(FOURCC_FMT);
    waveHdr.chkFmt.dwSize   = NSSwapHostLongToLittle(sizeof (PCMWAVEFORMAT));
    waveHdr.wFormatTag      = NSSwapHostShortToLittle(WAVE_FORMAT_PCM);
    waveHdr.nChannels       = NSSwapHostShortToLittle(channels);
    waveHdr.nSamplesPerSec  = NSSwapHostLongToLittle(rate);
    waveHdr.nBlockAlign     = NSSwapHostShortToLittle(nBlockAlign);
    waveHdr.nAvgBytesPerSec = NSSwapHostLongToLittle(nAvgBytesPerSec);
    waveHdr.wBitsPerSample  = NSSwapHostShortToLittle(nBitsPerSample);
    waveHdr.chkData.ckid    = NSSwapHostLongToLittle(FOURCC_DATA);
    waveHdr.chkData.dwSize  = NSSwapHostLongToLittle(byteCount);

    rc = write(fd, &waveHdr, sizeof(waveHdr));
    if (rc != sizeof(waveHdr)) {
        fprintf(stderr, "%s: Unable to write wav header.\n", pgm);
        perror("write");
        exit(1);
    }
}

void main (int argc, const char *argv[])
{
    NSAutoreleasePool  *pool = [[NSAutoreleasePool alloc] init];
    OSCSITarget        *target;
    OSCSIUnit          *unit;
    OSCSICDAudioStream *stream;
    unsigned int        startingLBA, lbaCount;
    int                 fd, rc;
    OSCSICDTableOfContents *toc;
    int originalUid = getuid(), originalGid = getgid();
    unsigned int        bytesPerReport, nextReportBytes, currentBytes, totalBytes;
    BOOL                swapToLittleEndian = NO;
 
    id <NSObject, OSCSICDRomAudioControl, OSCSICDRomAudioData> cdAudio;

    pgm = argv[0];
    [OBPostLoader processClasses];

    if (argc != 5 && argc != 6) {
        fprintf(stderr, "usage: %s scsi-raw-device file-format output-file starting-logical-block-address [logical-block-count]\n", argv[0]);
        exit(1);
    }

    startingLBA = atoi(argv[4]);

    NS_DURING{
        target = [[OSCSITarget alloc] initWithBusNumber: OSCSI_INVALID_BUS_NUMBER
                                           targetNumber: OSCSI_INVALID_TARGET_NUMBER
                                          rawDeviceName: [NSString stringWithCString: argv[1]]];
        unit = [target unitWithUnitNumber: 0];
        [unit open];
    } NS_HANDLER {
        NSLog(@"Error connecting to requested unit: %@", localException);
        exit(1);
    } NS_ENDHANDLER;

    if (![unit conformsToProtocol: @protocol(OSCSICDRomAudioControl)]) {
        fprintf(stderr, "The specified unit is not a CD Audio device.\n%s\n",
                [[[unit debugDictionary] description] cString]);
        exit(1);
    }
    if (![unit conformsToProtocol: @protocol(OSCSICDRomAudioData)]) {
        fprintf(stderr, "The specified unit is doesn't support CDDA reading.\n%s\n",
                [[[unit debugDictionary] description] cString]);
        exit(1);
    }

    printf("Selected unit: %s\n", [[[unit debugDictionary] description] cString]);

    cdAudio = (id <NSObject, OSCSICDRomAudioControl, OSCSICDRomAudioData>) unit;
    [cdAudio updateStatus];

    toc = [cdAudio tableOfContents];

    // Print out the toc
    printf("TOC: %s\n", [[[toc debugDictionary] description] cString]);

    if (argc == 6)
        lbaCount = atoi(argv[5]);
    else
        lbaCount = [toc startingLBAForTrackNumber: OSCSI_LEAD_OUT_TRACK] - startingLBA;

    totalBytes = lbaCount * OSCSI_CDDA_BYTES_PER_BLOCK;
    
    // Open the output file (create it as the real user)
    seteuid(originalUid);
    setegid(originalGid);
    fd = open(argv[3], O_CREAT|O_RDWR, 0660);
    if (fd < 0) {
        fprintf(stderr, "%s: Cannot open %s for writing.\n", argv[0], argv[3]);
        perror("open");
        exit(1);
    }
    seteuid(0);

    // Write the sound header
    if (strcmp(argv[2], "wav") == 0) {
        // wav must be little endian 
        swapToLittleEndian = YES;
        writeWAVHeader(fd, totalBytes);
    } else if (strcmp(argv[2], "snd") == 0)
        // snd files can be either endianness (based on the magic number)
        writeSNDHeader(fd, totalBytes);
    else {
        fprintf(stderr, "Unknown sound file format '%s'.  Must be wav or snd.\n", argv[2]);
        exit(1);
    }

    // Grab the requested audio and stick it in a .snd file.  Read several blocks at a time
    fprintf(stderr, "Reading from block %d to block %d (%d bytes).\n", startingLBA, startingLBA + lbaCount, totalBytes);

    stream = [[OSCSICDAudioStream alloc] initWithAudioSource: cdAudio
                                                 startingLBA: startingLBA
                                                   endingLBA: startingLBA + lbaCount
                                           blocksPerTransfer: 16];

    bytesPerReport = totalBytes/1000;
    currentBytes = 0;
    nextReportBytes = bytesPerReport;
    
    while (YES) {
        NSAutoreleasePool    *subPool;
        NSData               *data;
        unsigned int          length;
        const unsigned short *samples;
        
        subPool = [[NSAutoreleasePool alloc] init];
        data = [stream nextAudioBlock];
        if (!data)
            break;
        
        length = [data length];
        samples = [data bytes];

        if (swapToLittleEndian) {
#warning This is bad.  The data isn't necessarily mutable
            unsigned int sampleCount = length / 2;
            unsigned int sampleIndex;

            for (sampleIndex = 0; sampleIndex < sampleCount; sampleIndex++)
                samples[sampleIndex] = NSSwapHostShortToLittle(samples[sampleIndex]);
        }

        rc = write(fd, samples, length);
        if (rc != [data length]) {
            fprintf(stderr, "%s: Unable to write data to file.\n", argv[0]);
            perror("write");
            exit(1);
        }

        currentBytes += length;
        if (currentBytes > nextReportBytes) {
            nextReportBytes += bytesPerReport;
            fprintf(stderr, "%.1f%% done, %d bytes written ...\r", (currentBytes * 100.0) / totalBytes, currentBytes);
        }
        
        [subPool release];
    }
    fprintf(stderr, "%.1f%% done, %d bytes written ...\n", (currentBytes * 100.0) / totalBytes, currentBytes);

    [pool release];
    exit(0);       // insure the process exit status is 0
}
